from collections.abc import AsyncIterator
from typing import Final, Optional

from typing_extensions import Unpack

from ncae_sdk._async.endpoint import AsyncEndpoint
from ncae_sdk._resource import ResourceId
from ncae_sdk._util import map_async
from ncae_sdk.resources._schema import (
    SharedDataSubscriber,
    SharedDataSubscriberCreate,
    SharedDataSubscriberCreateModel,
    SharedDataSubscriberFilter,
    SharedDataSubscriberFilterModel,
    SharedDataSubscriberUpdate,
    SharedDataSubscriberUpdateModel,
)


class AsyncSharedDataSubscriberEndpoint(AsyncEndpoint):
    BASE_PATH: Final[str] = "shared_data/v1/subscriber"

    async def list(self, **filters: Unpack[SharedDataSubscriberFilter]) -> AsyncIterator[SharedDataSubscriber]:
        results = self._list(self.BASE_PATH, SharedDataSubscriberFilterModel, filters)
        return map_async(SharedDataSubscriber.parse_api, results)

    async def find(self, **filters: Unpack[SharedDataSubscriberFilter]) -> Optional[SharedDataSubscriber]:
        result = await self._get_by_filters(self.BASE_PATH, SharedDataSubscriberFilterModel, filters)
        return SharedDataSubscriber.parse_api(result) if result else None

    async def get(self, rid: ResourceId) -> Optional[SharedDataSubscriber]:
        result = await self._get_by_id(self.BASE_PATH, rid)
        return SharedDataSubscriber.parse_api(result) if result else None

    async def create(self, **payload: Unpack[SharedDataSubscriberCreate]) -> SharedDataSubscriber:
        result = await self._create(self.BASE_PATH, SharedDataSubscriberCreateModel, payload)
        return SharedDataSubscriber.parse_api(result)

    async def update(self, rid: ResourceId, /, **payload: Unpack[SharedDataSubscriberUpdate]) -> SharedDataSubscriber:
        result = await self._update(self.BASE_PATH, rid, SharedDataSubscriberUpdateModel, payload)
        return SharedDataSubscriber.parse_api(result)

    async def delete(self, rid: ResourceId) -> bool:
        return await self._delete(self.BASE_PATH, rid)



from ncae_sdk.fastapi._app import create_module_app
from ncae_sdk.fastapi._context import EmptyModel

__all__ = [
    # Re-exported classes and functions
    "create_module_app",
    "EmptyModel",
    # Submodules
    "extapi",
    "runner",
]



from typing import Any, Generic, Optional, Union

from pydantic import BaseModel, Field, GetJsonSchemaHandler, model_validator
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import CoreSchema
from typing_extensions import Self, TypeAlias, TypedDict, TypeVar

from ncae_sdk.fastapi._context import BaseRequestModel, BaseResponseModel, EmptyModel
from ncae_sdk.fastapi._utils import extract_pydantic_model_keys

CmdbT = TypeVar("CmdbT", bound=BaseModel, default=EmptyModel)
CallbackT = TypeVar("CallbackT", bound=BaseModel, default=EmptyModel)
QueryT = TypeVar("QueryT", bound=BaseModel, default=EmptyModel)

ExtApiReportResponseT = TypeVar("ExtApiReportResponseT", bound=BaseModel)


class ExtApiCredential(BaseRequestModel):
    name: str = Field(alias="name")
    username: str = Field(alias="username")
    password: str = Field(alias="password")
    become_password: str = Field(alias="become_password")


class ExtApiDeviceModel(BaseRequestModel):
    id: int = Field(alias="id")
    name: str = Field(alias="name")
    slug: str = Field(alias="slug")


class ExtApiDevice(BaseRequestModel):
    id: int = Field(alias="id")
    name: str = Field(alias="name")
    host: str = Field(alias="ip")
    verify_tls: bool = Field(alias="verify_tls")
    group_ids: list[int] = Field(alias="device_groups")
    target_names: list[str] = Field(alias="device_targets", default_factory=list)
    extra_vars: dict[str, Any] = Field(alias="extra_vars")
    device_model: ExtApiDeviceModel = Field(alias="device_model")
    credential: Optional[ExtApiCredential] = Field(alias="credential", default=None, validate_default=True)


class ExtApiPhaseExtraVars(BaseRequestModel, Generic[CmdbT, CallbackT]):
    is_decommission: bool = Field(alias="decommission")
    cmdb_data: CmdbT = Field(alias="data")
    cmdb_data_previous: Optional[CmdbT] = Field(alias="old_data", default=None)
    callback_data: CallbackT = Field(alias="callback_data", frozen=False)
    devices: list[ExtApiDevice] = Field(alias="devices")

    ncae_base_url: str = Field(alias="ncae_base_url")
    ncae_service_id: int = Field(alias="service_id")
    ncae_service_instance_id: int = Field(alias="service_instance_id")
    ncae_phase_instance_id: int = Field(alias="phase_instance_id")
    ncae_cmdb_entry_id: int = Field(alias="ncae_cmdb_entry_id")
    ncae_tenant_id: int = Field(alias="ncae_tenant_id")
    ncae_tenant_slug: str = Field(alias="ncae_tenant_slug")
    ncae_transaction_id: str = Field(alias="transaction_id")

    @model_validator(mode="before")
    @classmethod
    def inject_defaults(cls, data: Any) -> Any:
        # Due to `callback_data` being a generic, there is no easy way to inject an "empty" default value.
        # Therefore, this model validator ensures that at least an empty dict is provided if none was given.
        # Additionally, any non-dict value is forced to be an empty dict, as old NCAE Core versions might send `null`.
        if isinstance(data, dict):
            if not isinstance(data.get("callback_data"), dict):
                data["callback_data"] = {}

        return data


class ExtApiPhaseRequest(BaseRequestModel, Generic[CmdbT, CallbackT]):
    limit: Optional[str] = Field(alias="limit", default=None)
    extra_vars: ExtApiPhaseExtraVars[CmdbT, CallbackT] = Field(alias="extra_vars")


class ExtApiReportRequest(BaseRequestModel, Generic[QueryT]):
    devices: list[ExtApiDevice] = Field(alias="devices", default_factory=list)
    ncae_tenant_id: int = Field(alias="tenant")

    # Technically, all custom fields provided by a template are provided as top-level values.
    # To still provide some structure, these are hereby grouped under `query`, with the user-specified model.
    query: QueryT = Field(alias="query")

    # When an OTR is associated with a single device, its data will be provided top-level here.
    # The modern alternative is to use device / device groups targeting instead, stored in `devices`.
    device_host: Optional[str] = Field(alias="ip", default=None)
    device_verify_tls: Optional[bool] = Field(alias="verify_tls", default=None)
    device_username: Optional[str] = Field(alias="username", default=None)
    device_password: Optional[str] = Field(alias="password", default=None)
    device_become_password: Optional[str] = Field(alias="become_password", default=None)

    @model_validator(mode="before")
    @classmethod
    def validate_top_level_query(cls, values: dict[str, Any]) -> dict[str, Any]:
        if not isinstance(values, dict):
            return values

        # Extract all keys that do not belong to this model, respecting potential aliases
        own_keys = extract_pydantic_model_keys(cls)
        extra_keys = set(values.keys()).difference(own_keys)

        # Move all these extra keys under `query` to parse with the user-provided model
        query_data: dict[str, Any] = {}
        for extra_key in extra_keys:
            query_data[extra_key] = values.pop(extra_key)
        values["query"] = query_data

        return values

    @model_validator(mode="after")
    def validate_static_device(self) -> Self:
        base_fields = [self.device_host, self.device_verify_tls]
        credential_fields = [self.device_username, self.device_password, self.device_become_password]

        if any(f is not None for f in base_fields + credential_fields) and not all(f is not None for f in base_fields):
            raise ValueError("If any device field is provided, all base fields must be provided")

        if any(f is not None for f in credential_fields) and not all(f is not None for f in credential_fields):
            raise ValueError("If any device credential field is provided, all credential fields must be provided")

        return self

    @model_validator(mode="after")
    def validate_query_model(self) -> Self:
        own_fields = extract_pydantic_model_keys(self.__class__)
        query_fields = extract_pydantic_model_keys(self.query.__class__)
        overlapping_fields = own_fields.intersection(query_fields)

        if overlapping_fields:
            raise ValueError(f"Query model fields overlap with request fields: {', '.join(overlapping_fields)}")

        return self

    @classmethod
    def __get_pydantic_json_schema__(cls, core_schema: CoreSchema, handler: GetJsonSchemaHandler, /) -> JsonSchemaValue:
        schema = handler(core_schema)
        schema = handler.resolve_ref_schema(schema)
        assert schema.get("type") == "object" and isinstance(schema.get("properties"), dict)

        query_schema = schema["properties"].pop("query")
        query_schema = handler.resolve_ref_schema(query_schema)
        assert query_schema.get("type") == "object" and isinstance(query_schema.get("properties"), dict)

        schema["properties"].update(query_schema["properties"])
        return schema


ExtApiReportFieldValue: TypeAlias = Union[str, int, float, bool, None]


class ExtApiReportLinkField(TypedDict):
    value: ExtApiReportFieldValue
    href: str


class ExtApiReportHtmlField(TypedDict):
    value: ExtApiReportFieldValue
    extension: str


ExtApiReportField: TypeAlias = Union[
    ExtApiReportFieldValue,
    ExtApiReportLinkField,
    ExtApiReportHtmlField,
]


class ExtApiReportResponse(BaseResponseModel):
    field_names: list[str] = Field(serialization_alias="fields")
    field_data: list[dict[str, ExtApiReportField]] = Field(serialization_alias="field_data")



# @generated
# This file was automatically generated by the ncae-sdk code generator.
# DO NOT EDIT MANUALLY, ANY CHANGES WILL BE OVERWRITTEN.

from collections.abc import Iterator
from typing import Final, Optional

from typing_extensions import Unpack

from ncae_sdk._resource import ResourceId
from ncae_sdk._sync.endpoint import Endpoint
from ncae_sdk._util import map_sync
from ncae_sdk.resources._schema import (
    Tag,
    TagCreate,
    TagCreateModel,
    TagFilter,
    TagFilterModel,
    TagUpdate,
    TagUpdateModel,
)


class TagEndpoint(Endpoint):
    BASE_PATH: Final[str] = "tag/v1/tag"

    def list(self, **filters: Unpack[TagFilter]) -> Iterator[Tag]:
        results = self._list(self.BASE_PATH, TagFilterModel, filters)
        return map_sync(Tag.parse_api, results)

    def find(self, **filters: Unpack[TagFilter]) -> Optional[Tag]:
        result = self._get_by_filters(self.BASE_PATH, TagFilterModel, filters)
        return Tag.parse_api(result) if result else None

    def get(self, rid: ResourceId) -> Optional[Tag]:
        result = self._get_by_id(self.BASE_PATH, rid)
        return Tag.parse_api(result) if result else None

    def create(self, **payload: Unpack[TagCreate]) -> Tag:
        result = self._create(self.BASE_PATH, TagCreateModel, payload)
        return Tag.parse_api(result)

    def update(self, rid: ResourceId, /, **payload: Unpack[TagUpdate]) -> Tag:
        result = self._update(self.BASE_PATH, rid, TagUpdateModel, payload)
        return Tag.parse_api(result)

    def delete(self, rid: ResourceId) -> bool:
        return self._delete(self.BASE_PATH, rid)



from collections.abc import AsyncIterator
from typing import Final, Optional, cast

from typing_extensions import Unpack

from ncae_sdk._async.endpoint import AsyncEndpoint
from ncae_sdk._resource import ResourceId
from ncae_sdk._util import map_async
from ncae_sdk.resources._schema import (
    Service,
    ServiceCreate,
    ServiceCreateModel,
    ServiceFilter,
    ServiceFilterModel,
    ServiceUpdate,
    ServiceUpdateModel,
)
from ncae_sdk.types import ServiceConfig


class AsyncServiceEndpoint(AsyncEndpoint):
    BASE_PATH: Final[str] = "automation/v1/service"

    async def list(self, **filters: Unpack[ServiceFilter]) -> AsyncIterator[Service]:
        results = self._list(self.BASE_PATH, ServiceFilterModel, filters)
        return map_async(Service.parse_api, results)

    async def find(self, **filters: Unpack[ServiceFilter]) -> Optional[Service]:
        result = await self._get_by_filters(self.BASE_PATH, ServiceFilterModel, filters)
        return Service.parse_api(result) if result else None

    async def get(self, rid: ResourceId) -> Optional[Service]:
        result = await self._get_by_id(self.BASE_PATH, rid)
        return Service.parse_api(result) if result else None

    async def create(self, **payload: Unpack[ServiceCreate]) -> Service:
        result = await self._create(self.BASE_PATH, ServiceCreateModel, payload)
        return Service.parse_api(result)

    async def update(self, rid: ResourceId, /, **payload: Unpack[ServiceUpdate]) -> Service:
        result = await self._update(self.BASE_PATH, rid, ServiceUpdateModel, payload)
        return Service.parse_api(result)

    async def delete(self, rid: ResourceId) -> bool:
        return await self._delete(self.BASE_PATH, rid)

    async def export_config(self, rid: ResourceId) -> ServiceConfig:
        """
        Exports the specified NCAE service configuration with all its phases as a portable JSON object.
        This exported configuration can be used for backup purposes or to import the service into another NCAE instance.
        """
        response = await self._session.get(f"{self.BASE_PATH}/{rid}/export")
        return cast(ServiceConfig, response.json())

    async def import_config(self, data: ServiceConfig) -> Service:
        """
        Imports a service configuration into NCAE from a JSON object that was previously exported using export_config.
        The provided configuration should include all necessary details for the service and its phases.
        If a service with the same name already exists, it will be updated; otherwise, a new service is created.
        """

        result = await self._session.post(f"{self.BASE_PATH}/import", json=data)
        service = await self.get(result.json()["id"])
        assert service is not None, "Service should exist after import"
        return service



import logging

from fastapi import Request
from fastapi.exception_handlers import request_validation_exception_handler
from fastapi.exceptions import RequestValidationError
from starlette.responses import JSONResponse

from ncae_sdk.fastapi._utils import obtain_request_logger

default_logger = logging.getLogger("ncae_sdk")


async def validation_exception_handler(request: Request, exc: RequestValidationError) -> JSONResponse:
    logger = obtain_request_logger(request)
    for error in exc.errors():
        type_ = error.get("type", "unknown")
        message = error.get("msg", "Unknown validation error")
        location = ".".join(str(part) for part in error.get("loc", []))

        if location:
            logger.error("Validation failed at '%s': %s (type: %s)", location, message, type_)
        else:
            logger.error("Validation failed: %s (type: %s)", message, type_)

    return await request_validation_exception_handler(request, exc)



# @generated
# This file was automatically generated by the ncae-sdk code generator.
# DO NOT EDIT MANUALLY, ANY CHANGES WILL BE OVERWRITTEN.

from typing import Final, Iterator, Optional

from typing_extensions import Unpack

from ncae_sdk._resource import ResourceId
from ncae_sdk._sync.endpoint import Endpoint
from ncae_sdk._util import map_sync
from ncae_sdk.resources._schema import (
    Report,
    ReportCreate,
    ReportCreateModel,
    ReportFilter,
    ReportFilterModel,
    ReportUpdate,
    ReportUpdateModel,
)


class ReportEndpoint(Endpoint):
    BASE_PATH: Final[str] = "dashboard/v1/onetimereport"

    def list(self, **filters: Unpack[ReportFilter]) -> Iterator[Report]:
        results = self._list(self.BASE_PATH, ReportFilterModel, filters)
        return map_sync(Report.parse_api, results)

    def find(self, **filters: Unpack[ReportFilter]) -> Optional[Report]:
        result = self._get_by_filters(self.BASE_PATH, ReportFilterModel, filters)
        return Report.parse_api(result) if result else None

    def get(self, rid: ResourceId) -> Optional[Report]:
        result = self._get_by_id(self.BASE_PATH, rid)
        return Report.parse_api(result) if result else None

    def create(self, **payload: Unpack[ReportCreate]) -> Report:
        result = self._create(self.BASE_PATH, ReportCreateModel, payload)
        return Report.parse_api(result)

    def update(self, rid: ResourceId, /, **payload: Unpack[ReportUpdate]) -> Report:
        result = self._update(self.BASE_PATH, rid, ReportUpdateModel, payload)
        return Report.parse_api(result)

    def delete(self, rid: ResourceId) -> bool:
        return self._delete(self.BASE_PATH, rid)



# @generated
# This file was automatically generated by the ncae-sdk code generator.
# DO NOT EDIT MANUALLY, ANY CHANGES WILL BE OVERWRITTEN.

from collections.abc import Iterator
from typing import Any, Final

from typing_extensions import Unpack

from ncae_sdk._http import Response
from ncae_sdk._sync.endpoint import Endpoint
from ncae_sdk._util import map_sync, next_sync
from ncae_sdk.resources._schema import (
    TagRelation,
    TagRelationCreate,
    TagRelationCreateModel,
    TagRelationFilter,
    TagRelationFilterModel,
)


class TagRelationEndpoint(Endpoint):
    BASE_PATH: Final[str] = "tag/v1/tag-relations"

    def list(self, **filters: Unpack[TagRelationFilter]) -> Iterator[TagRelation]:
        results = self._list(
            self.BASE_PATH,
            TagRelationFilterModel,
            filters,
            paginate=False,
            postprocessor=self._flatten_response,
        )
        return map_sync(TagRelation.parse_api, results)

    def set(self, **payload: Unpack[TagRelationCreate]) -> TagRelation:
        self._create(self.BASE_PATH, TagRelationCreateModel, payload)
        results = self.list(object_type=payload["object_type"], object_ids=[payload["object_id"]])
        relation = next_sync(results)
        if relation is None:
            raise ValueError("No relation found after creation")

        return relation

    @classmethod
    def _flatten_response(cls, response: Response) -> Any:
        """
        The API response for this endpoint is somewhat exotic, given it returns a dictionary of object id to tag ids.
        To match our overall expected data structure, flatten the response into a list of dictionaries.
        Additionally, include the object type in each dictionary for clarity, which can be extracted from the request.
        """

        object_type = response.request.query_params["model_name"]
        response_data = response.json()
        if not isinstance(response_data, dict):
            raise ValueError("Invalid response data format")

        results = []
        for object_id, tags in response_data.items():
            if not isinstance(tags, list):
                raise ValueError(f"Invalid tag IDs format for object ID {object_id}")

            tag_ids: list[int] = [int(tag["id"]) for tag in tags]
            results.append({"object_type": object_type, "object_id": object_id, "tag_ids": tag_ids})

        return results



from __future__ import annotations

from typing import Any, Final, Optional

from jinja2 import Environment
from jinja2.nodes import Call, Const, For, Getattr, Getitem, Name, Node, Tuple
from jinja2.visitor import NodeVisitor
from typing_extensions import TypeAlias

VariablePath: TypeAlias = tuple[str, ...]  # e.g. (ntp_servers, ip) implies ntp_servers.ip
IterablePath: TypeAlias = tuple[str, ...]  # e.g. (users,params) implies a nested loop for users: for params: x
AliasMap: TypeAlias = dict[str, list[str]]  # e.g. loop var name -> iterable path, e.g. {"param": ["users", "params"]}

DYNAMIC_KEY_PLACEHOLDER: Final[str] = "<key>"  # placeholder for non-constant/unknown index keys


# noinspection PyPep8Naming
class VariableCollector(NodeVisitor):
    """
    Collects all variable paths and iterable roots from a Jinja2 AST.
    - Uses a stack to detect when a `Name` node is part of a chain.
    - Tracks loop aliases to resolve variable paths within loops
    - Returns sets of absolute paths for both variables and iterables.
    """

    def __init__(self) -> None:
        self.paths: set[VariablePath] = set()
        self.iterables: set[IterablePath] = set()
        self._node_stack: Final[list[Node]] = []
        self._alias_stack: list[AliasMap] = []

    def visit(self, node: Node, *args: Any, **kwargs: Any) -> None:
        """
        Visit a node while tracking the parent on a simple stack.
        """
        self._node_stack.append(node)
        super().visit(node)
        self._node_stack.pop()

    def visit_Name(self, node: Name) -> None:
        """
        Visit each Name node, which is used in Jinja to lookup or store a value in a name.
        Any flat variable like "{{ abc }}" will be represented as a Name node.
        """

        # Ignore any non-load contexts, e.g. function params or assignments.
        if node.ctx != "load":
            return

        # Ignore if part of a chained access, e.g. abc.def or abc[def], otherwise this would double-count.
        parent = self._get_node_parent()
        if isinstance(parent, (Getattr, Getitem)):
            return

        # Resolve alias (if any) and record the absolute path
        path = self._resolve_alias([node.name])
        self.paths.add(tuple(path))

    def visit_Getattr(self, node: Getattr) -> None:
        """
        Visit each Getattr node, which represents attribute access in Jinja.
        There is some special case handling to avoid recording paths for calls to items/keys/values.
        E.g. "{{ foo.bar }}" is represented as Getattr(node=Name(foo), attr=bar).
        """
        # Detect if this is a call to items/keys/values
        parent = self._get_node_parent()
        is_callee = isinstance(parent, Call) and parent.node is node and node.attr in {"items", "keys", "values"}

        # Skip recording this path if it's the callee of an items/keys/values call.
        if not is_callee:
            path = self._expr_to_path(node)
            if path is not None:
                self.paths.add(tuple(path))

        # Call default visitor which traverses child nodes.
        self.generic_visit(node)

    def visit_Getitem(self, node: Getitem) -> None:
        """
        Visit each Getitem node, which represents index access in Jinja.
        E.g. "{{ foo['bar'] }}" is represented as Getitem(node=Name(foo), arg=Const('bar')).
        """
        path = self._expr_to_path(node)
        if path is not None:
            self.paths.add(tuple(path))

        # Call default visitor which traverses child nodes.
        self.generic_visit(node)

    def visit_For(self, node: For) -> None:
        """
        Visit each For node, which represents a for-loop in Jinja, e.g. "{% for u in users %}...{% endfor %}".
        1. Collect the iterable path (e.g. "users").
        2. Push an alias mapping for the loop target(s) to the iterable path.
        3. Visit the loop body with the alias in scope.
        4. Visit the else body if present.
        5. Pop the alias mapping.
        """
        # Determine the iterable path, e.g. "users" in "for u in users".
        iterable_path = self._iterable_path_from_iter(node.iter)
        if iterable_path:
            self.iterables.add(tuple(iterable_path))
            self.paths.add(tuple(iterable_path))

        # Visit the iterable expression first to collect any nested variables.
        self.visit(node.iter)

        # Determine loop target names, e.g. "u" in "for u in users".
        targets = self._target_names(node.target)

        # Push a temporary alias scope for the loop targets before visiting the body.
        pushed = False
        if targets and iterable_path is not None:
            # Use special handling for dict-based loops with .items().
            # The key alias is being mapped to iterable/<key-alias> for clarity.
            # The remaining aliases (= the object values) map to the iterable path directly.
            alias_scope: AliasMap = {}
            if self._is_items_call(node.iter) and len(targets) >= 2:
                alias_scope = {
                    targets[0]: list(iterable_path) + [targets[0]],
                    **{target: list(iterable_path) for target in targets[1:]},
                }
            else:
                alias_scope = {target: list(iterable_path) for target in targets}

            self._alias_stack.append(alias_scope)
            pushed = True

        # Visit the loop body, and then the else body if present.
        for child in node.body:
            self.visit(child)
        else_body = getattr(node, "else_", []) or []
        for child in else_body:
            self.visit(child)

        # Remove the alias scope after visiting the loop.
        if pushed:
            self._alias_stack.pop()

    def _is_alias_name(self, name: str) -> bool:
        """
        Returns true if the name is currently bound as a loop alias.
        E.g. in "{% for u in users %} {{ u.name }} {% endfor %}", "u" is an alias for "users".
        """
        for scope in reversed(self._alias_stack):
            if name in scope:
                return True

        return False

    def _resolve_alias(self, path: list[str]) -> list[str]:
        """
        Resolves the first element of the path if it is a loop alias.
        E.g. in "{% for u in users %} {{ u.name }} {% endfor %}", "u.name" resolves to "users.name".
        """
        if not path:
            return path

        head = path[0]
        for scope in reversed(self._alias_stack):
            if head in scope:
                return list(scope[head]) + path[1:]

        return path

    def _get_node_parent(self) -> Optional[Node]:
        """
        Get the parent node of the current node on the stack, or None if at root.
        """
        if len(self._node_stack) >= 2:
            return self._node_stack[-2]

        return None

    def _expr_to_path(self, node: Node) -> Optional[list[str]]:
        """
        Convert an attribute / index chain to a path, resolving the base alias if needed.

        Examples:
        - foo.bar => ["foo", "bar"]
        - foo["bar"].baz => ["foo", "bar", "baz"]
        - foo[i] => ["foo", "<key>"]
        """
        segments: list[str] = []
        current: Optional[Node] = node

        while isinstance(current, (Getattr, Getitem)):
            if isinstance(current, Getattr):
                segments.append(current.attr)
                current = current.node
            else:
                arg = current.arg
                if isinstance(arg, Const) and isinstance(arg.value, str):
                    segments.append(arg.value)
                else:
                    segments.append(DYNAMIC_KEY_PLACEHOLDER)
                current = current.node

        if isinstance(current, Name):
            base = self._resolve_alias([current.name])
            return base + list(reversed(segments))

        return None

    def _iterable_path_from_iter(self, node: Node) -> Optional[list[str]]:
        if self._is_items_call(node):
            assert isinstance(node, Call)
            assert isinstance(node.node, Getattr)
            return self._expr_to_path(node.node.node)

        return self._expr_to_path(node)

    @staticmethod
    def _is_items_call(node: Node) -> bool:
        return isinstance(node, Call) and isinstance(node.node, Getattr) and node.node.attr == "items"

    @staticmethod
    def _target_names(target: Node) -> list[str]:
        """
        Extract loop target variable names, both single (for x in y) and unpacked (for x, y in z).
        """
        if isinstance(target, Name):
            return [target.name]
        if isinstance(target, Tuple):
            return [node.name for node in target.items if isinstance(node, Name)]

        try:
            return [node.name for node in target.find_all(Name)]
        except Exception:
            return []


def collect_variables(jinja_template: str) -> tuple[set[VariablePath], set[IterablePath]]:
    jinja_env = Environment()
    ast = jinja_env.parse(jinja_template)
    collector = VariableCollector()
    collector.visit(ast)

    return collector.paths, collector.iterables



# @generated
# This file was automatically generated by the ncae-sdk code generator.
# DO NOT EDIT MANUALLY, ANY CHANGES WILL BE OVERWRITTEN.

from collections.abc import Iterator
from contextlib import contextmanager
from typing import IO, Final, Optional

from typing_extensions import Unpack, deprecated

from ncae_sdk._resource import ResourceId
from ncae_sdk._sync.endpoint import Endpoint
from ncae_sdk._util import map_sync
from ncae_sdk.resources._schema import (
    UploadFile,
    UploadFileCreate,
    UploadFileCreateModel,
    UploadFileFilter,
    UploadFileFilterModel,
    UploadFileUpdate,
    UploadFileUpdateModel,
)


class UploadFileEndpoint(Endpoint):
    BASE_PATH: Final[str] = "automation/v1/upload-file"

    def list(self, **filters: Unpack[UploadFileFilter]) -> Iterator[UploadFile]:
        results = self._list(self.BASE_PATH, UploadFileFilterModel, filters)
        return map_sync(UploadFile.parse_api, results)

    def create(self, file: IO[bytes], **payload: Unpack[UploadFileCreate]) -> UploadFile:
        result = self._create_multipart(self.BASE_PATH, UploadFileCreateModel, payload, {"file": file})
        return UploadFile.parse_api(result)

    def find(self, **filters: Unpack[UploadFileFilter]) -> Optional[UploadFile]:
        result = self._get_by_filters(self.BASE_PATH, UploadFileFilterModel, filters)
        return UploadFile.parse_api(result) if result else None

    def get(self, rid: ResourceId) -> Optional[UploadFile]:
        result = self._get_by_id(self.BASE_PATH, rid)
        return UploadFile.parse_api(result) if result else None

    def update(self, rid: ResourceId, /, **payload: Unpack[UploadFileUpdate]) -> UploadFile:
        result = self._update(self.BASE_PATH, rid, UploadFileUpdateModel, payload)
        return UploadFile.parse_api(result)

    def delete(self, rid: ResourceId) -> bool:
        return self._delete(self.BASE_PATH, rid)

    @deprecated("Use 'create' method instead, which is more appropriately named and has the same signature")
    def upload(self, file: IO[bytes], **payload: Unpack[UploadFileCreate]) -> UploadFile:
        return self.create(file, **payload)

    @contextmanager
    def download(self, rid: ResourceId, *, chunk_size: Optional[int] = None) -> Iterator[Iterator[bytes]]:
        with self._session.stream("GET", f"{self.BASE_PATH}/{rid}/download") as response:
            yield response.iter_bytes(chunk_size=chunk_size)



from __future__ import annotations

from functools import lru_cache

from pydantic import AliasChoices, Field
from pydantic_settings import BaseSettings, SettingsConfigDict


def optional_prefix(name: str) -> AliasChoices:
    return AliasChoices(f"extapi_{name}", name)


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        env_file=".env",
        env_file_encoding="utf-8",
        env_prefix="EXTAPI_",
        extra="ignore",
    )

    debug: bool = False

    # The following fields are preferably set with the "EXTAPI_" prefix, but can also be set without it for convenience.
    # This is especially useful as various older deployments tend to use the non-prefixed environment variables.
    ncae_base_url: str = Field(validation_alias=optional_prefix("ncae_base_url"))
    ncae_username: str = Field(validation_alias=optional_prefix("ncae_username"))
    ncae_password: str = Field(validation_alias=optional_prefix("ncae_password"))
    ncae_verify_tls: bool = Field(default=False, validation_alias=optional_prefix("ncae_verify_tls"))


@lru_cache
def get_settings() -> Settings:
    return Settings()



import inspect
from concurrent.futures import ThreadPoolExecutor
from typing import Awaitable, Callable, Final, Union, cast

from fastapi import Request
from pydantic import BaseModel, ConfigDict
from starlette.concurrency import run_in_threadpool

from ncae_sdk._async.client import AsyncClient
from ncae_sdk._session import SessionContext
from ncae_sdk._sync.client import Client
from ncae_sdk._types import FuncP, FuncR
from ncae_sdk.fastapi._dependencies import AsyncNcaeClientDep, NcaeClientDep, ThreadPoolExecutorDep
from ncae_sdk.fastapi._logging import ContextLogger
from ncae_sdk.fastapi._utils import obtain_request_logger


class EmptyModel(BaseModel):
    model_config = ConfigDict(extra="allow")


class BaseRequestModel(BaseModel):
    model_config = ConfigDict(
        frozen=True,
        validate_by_name=False,
        validate_by_alias=True,
    )


class BaseResponseModel(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
        serialize_by_alias=True,
        validate_by_name=True,
        validate_by_alias=False,
    )


class BaseContext:
    def __init__(
        self,
        request: Request,
        ncae_client: NcaeClientDep,
        async_ncae_client: AsyncNcaeClientDep,
        thread_pool: ThreadPoolExecutorDep,
    ):
        self._request: Final[Request] = request
        self._context: Final[SessionContext] = self._build_session_context()
        self._internal_logger: Final[ContextLogger] = obtain_request_logger(request)

        self._client: Final[Client] = ncae_client.clone_with_context(self._context)
        self._async_client: Final[AsyncClient] = async_ncae_client.clone_with_context(self._context)
        self._thread_pool: Final[ThreadPoolExecutor] = thread_pool

    @property
    def client(self) -> Client:
        return self._client

    @property
    def async_client(self) -> AsyncClient:
        return self._async_client

    @property
    def context(self) -> SessionContext:
        return self._context

    @property
    def internal_logger(self) -> ContextLogger:
        return self._internal_logger

    @property
    def thread_pool(self) -> ThreadPoolExecutor:
        return self._thread_pool

    def _build_session_context(self) -> SessionContext:
        return SessionContext()


async def _run_async_or_sync_function(
    func: Callable[FuncP, Union[FuncR, Awaitable[FuncR]]],
    /,
    *func_args: FuncP.args,
    **func_kwargs: FuncP.kwargs,
) -> FuncR:
    if inspect.iscoroutinefunction(func):
        async_func = cast(Callable[FuncP, Awaitable[FuncR]], func)
        return await async_func(*func_args, **func_kwargs)
    else:
        sync_func = cast(Callable[FuncP, FuncR], func)
        return await run_in_threadpool(sync_func, *func_args, **func_kwargs)



from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import IO, Final, Optional

from typing_extensions import Unpack, deprecated

from ncae_sdk._async.endpoint import AsyncEndpoint
from ncae_sdk._resource import ResourceId
from ncae_sdk._util import map_async
from ncae_sdk.resources._schema import (
    UploadFile,
    UploadFileCreate,
    UploadFileCreateModel,
    UploadFileFilter,
    UploadFileFilterModel,
    UploadFileUpdate,
    UploadFileUpdateModel,
)


class AsyncUploadFileEndpoint(AsyncEndpoint):
    BASE_PATH: Final[str] = "automation/v1/upload-file"

    async def list(self, **filters: Unpack[UploadFileFilter]) -> AsyncIterator[UploadFile]:
        results = self._list(self.BASE_PATH, UploadFileFilterModel, filters)
        return map_async(UploadFile.parse_api, results)

    async def create(self, file: IO[bytes], **payload: Unpack[UploadFileCreate]) -> UploadFile:
        result = await self._create_multipart(self.BASE_PATH, UploadFileCreateModel, payload, {"file": file})
        return UploadFile.parse_api(result)

    async def find(self, **filters: Unpack[UploadFileFilter]) -> Optional[UploadFile]:
        result = await self._get_by_filters(self.BASE_PATH, UploadFileFilterModel, filters)
        return UploadFile.parse_api(result) if result else None

    async def get(self, rid: ResourceId) -> Optional[UploadFile]:
        result = await self._get_by_id(self.BASE_PATH, rid)
        return UploadFile.parse_api(result) if result else None

    async def update(self, rid: ResourceId, /, **payload: Unpack[UploadFileUpdate]) -> UploadFile:
        result = await self._update(self.BASE_PATH, rid, UploadFileUpdateModel, payload)
        return UploadFile.parse_api(result)

    async def delete(self, rid: ResourceId) -> bool:
        return await self._delete(self.BASE_PATH, rid)

    @deprecated("Use 'create' method instead, which is more appropriately named and has the same signature")
    async def upload(self, file: IO[bytes], **payload: Unpack[UploadFileCreate]) -> UploadFile:
        return await self.create(file, **payload)

    @asynccontextmanager
    async def download(
        self, rid: ResourceId, *, chunk_size: Optional[int] = None
    ) -> AsyncIterator[AsyncIterator[bytes]]:
        async with self._session.stream("GET", f"{self.BASE_PATH}/{rid}/download") as response:
            yield await response.aiter_bytes(chunk_size=chunk_size)



from collections.abc import AsyncIterator
from typing import Final

from typing_extensions import Unpack

from ncae_sdk._async.endpoint import AsyncEndpoint
from ncae_sdk._util import map_async
from ncae_sdk.resources._schema import (
    UsageLog,
    UsageLogFilter,
    UsageLogFilterModel,
)


class AsyncUsageLogEndpoint(AsyncEndpoint):
    BASE_PATH: Final[str] = "logger/v1/usage-log"

    async def list(self, **filters: Unpack[UsageLogFilter]) -> AsyncIterator[UsageLog]:
        results = self._list(self.BASE_PATH, UsageLogFilterModel, filters)
        return map_async(UsageLog.parse_api, results)



import functools
import inspect
from dataclasses import dataclass
from typing import Annotated, Any, Awaitable, Iterator, Optional

from typing_extensions import Callable, ParamSpec, TypeAlias, TypeVar
from typing_extensions import get_args as get_type_args
from typing_extensions import get_origin as get_type_origin

FuncP = ParamSpec("FuncP")
FuncR = TypeVar("FuncR")
Func = Callable[FuncP, FuncR]


@dataclass(frozen=True)
class ParamProcessorContext:
    current: list[inspect.Parameter]
    original: list[inspect.Parameter]

    @property
    def current_map(self) -> dict[str, inspect.Parameter]:
        return {param.name: param for param in self.current}

    @property
    def original_map(self) -> dict[str, inspect.Parameter]:
        return {param.name: param for param in self.original}


ParamProcessor: TypeAlias = Callable[[ParamProcessorContext], Iterator[inspect.Parameter]]


def process_function_parameters(
    *processors: ParamProcessor,
) -> Callable[[Callable[FuncP, Awaitable[FuncR]]], Callable[FuncP, Awaitable[FuncR]]]:
    def decorator(func: Callable[FuncP, Awaitable[FuncR]]) -> Callable[FuncP, Awaitable[FuncR]]:
        @functools.wraps(func)
        async def wrapper(*args: FuncP.args, **kwargs: FuncP.kwargs) -> FuncR:
            return await func(*args, **kwargs)

        # Obtain function signature and parameters
        original_signature = inspect.signature(func)
        original_params = list(original_signature.parameters.values())

        # Apply all processors in sequence to transform the function signature
        current_params = original_params
        for processor in processors:
            # Construct context and run parameter processor
            context = ParamProcessorContext(
                current=current_params,
                original=original_params,
            )
            current_params = list(processor(context))

            # Add any new parameters to original parameters for subsequent processors
            missing_params = [p for p in current_params if p.name not in context.original_map]
            if missing_params:
                original_params.extend(missing_params)

        # Replace wrapper signature with transformed signature
        wrapper_signature = original_signature.replace(parameters=current_params)
        setattr(wrapper, "__signature__", wrapper_signature)

        return wrapper

    return decorator


def inject_parameter(name: str, typ: type) -> ParamProcessor:
    def processor(ctx: ParamProcessorContext) -> Iterator[inspect.Parameter]:
        yield from ctx.current
        yield inspect.Parameter(
            name=name,
            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            annotation=typ,
        )

    return processor


def resolve_generic_signatures(ctx: ParamProcessorContext) -> Iterator[inspect.Parameter]:
    for param in ctx.current:
        original_param = ctx.original_map[param.name]
        param_class = _extract_class_type(original_param.annotation)

        if param_class:
            new_type = _resolve_generic_class_signature(original_param.annotation)
            new_annotation = _replace_annotation_type(param.annotation, new_type)
            yield param.replace(annotation=new_annotation)
        else:
            yield param


def _extract_class_type(annotation: type) -> Optional[type]:
    if inspect.isclass(annotation):
        return annotation
    elif origin := get_type_origin(annotation):
        if inspect.isclass(origin):
            return origin

    return None


def _replace_annotation_type(annotation: type, new_type: type) -> Any:
    origin = get_type_origin(annotation)
    args = get_type_args(annotation)

    if origin is Annotated and len(args) == 2:
        return Annotated[new_type, args[1]]
    else:
        return new_type


def _resolve_generic_class_signature(cls_annotation: type) -> Any:
    # Extract actual class (origin) with its generics and their arguments
    cls = _extract_class_type(cls_annotation)
    cls_args = get_type_args(cls_annotation)
    cls_generics = getattr(cls, "__parameters__", [])

    # Ensure type origin is actually a class and return early if there are no generics to resolve
    if not inspect.isclass(cls):
        raise TypeError("Provided argument is not a class.")
    if not cls_generics or not cls_args:
        return cls_annotation

    # Fill generic arguments with their defaults if not explicitly provided
    if len(cls_generics) != len(cls_args):
        filled_args = list(cls_args)
        for gen in cls_generics[len(cls_args) :]:
            if defaults := getattr(gen, "__default__", ()):
                filled_args.append(defaults)
        cls_args = tuple(filled_args)

    # Build mapping of generic type variables to their actual arguments
    if len(cls_generics) != len(cls_args):
        raise TypeError("Unable to resolve generics in class signature due to mismatched arguments.")
    cls_generic_map = {gen: arg for gen, arg in zip(cls_generics, cls_args)}

    # Generate wrapper function around class, so that its constructor signature can be modified
    @functools.wraps(cls)
    def cls_wrapper(*args: Any, **kwargs: Any) -> Any:
        return cls(*args, **kwargs)

    # Obtain the __init__ signature and parameters, skipping 'self'
    # The latter is implicitly handled by Python and must not be specified
    init_signature = inspect.signature(cls.__init__)
    init_params = list(init_signature.parameters.values())[1:]

    # Build new function parameters, resolving all generic variables where possible
    wrapper_params: list[inspect.Parameter] = []
    for init_param in init_params:
        # Forward existing parameter as-is if not originating from current package
        if not _is_from_current_package(init_param.annotation):
            wrapper_params.append(init_param)
            continue

        # Forward existing parameter as-is if there are no generics to resolve
        param_generics = getattr(init_param.annotation, "__parameters__", [])
        if not param_generics:
            wrapper_params.append(init_param)
            continue

        # Ensure parameter annotation points to class type
        if not inspect.isclass(init_param.annotation):
            raise TypeError(f"Parameter {init_param.name} annotation is not a class type.")

        # Resolve all generic variables in parameter annotation
        wrapper_param_args: list[type] = []
        for param_generic in param_generics:
            if concrete_type := cls_generic_map.get(param_generic, None):
                wrapper_param_args.append(concrete_type)
            else:
                raise TypeError(f"Unable to resolve generic {param_generic} in parameter {init_param.name}.")

        # Construct new parameter annotation with resolved generics
        wrapper_param_annotation = init_param.annotation[tuple(wrapper_param_args)]
        wrapper_param = init_param.replace(annotation=wrapper_param_annotation)
        wrapper_params.append(wrapper_param)

    # Replace wrapper signature with resolved parameters
    wrapper_signature = init_signature.replace(parameters=wrapper_params)
    setattr(cls_wrapper, "__signature__", wrapper_signature)

    return cls_wrapper


def _is_from_current_package(type_: Any, *, top_level_pkg: Optional[str] = None, depth: int = 0) -> bool:
    # Limit recursion depth to prevent infinite loops
    if depth > 5:
        return False

    # Determine top-level package for comparison
    if not top_level_pkg:
        pkg = __package__ or __name__
        if not pkg or pkg == "__main__":
            top_level_pkg = "ncae_sdk"
        else:
            top_level_pkg = pkg.split(".", 1)[0]

    # If type has arguments, check their origin as well to handle cases like list[X] or Annotated[X, ...]
    type_args = get_type_args(type_)
    for arg in type_args or []:
        if _is_from_current_package(
            arg,
            top_level_pkg=top_level_pkg,
            depth=depth + 1,
        ):
            return True

    # Resolve module path for type, considering as foreign if no module information is available
    module_path = getattr(type_, "__module__", None)
    if not module_path:
        return False

    # Walk up module path to determine top-level package
    while module_path:
        if module_path == top_level_pkg:
            return True
        if "." not in module_path:
            break
        module_path = module_path.rsplit(".", 1)[0]

    return False



from ncae_sdk.__version__ import __description__, __title__, __version__
from ncae_sdk._async.client import AsyncClient
from ncae_sdk._async.endpoint import AsyncEndpoint
from ncae_sdk._auth import Auth, SessionAuth
from ncae_sdk._resource import Resource, ResourceId, Schema
from ncae_sdk._session import SessionContext
from ncae_sdk._sync.client import Client
from ncae_sdk._sync.endpoint import Endpoint

__all__ = [
    # Re-exported classes and functions
    "AsyncClient",
    "AsyncEndpoint",
    "Auth",
    "Client",
    "Endpoint",
    "Resource",
    "ResourceId",
    "Schema",
    "SessionAuth",
    "SessionContext",
    "__description__",
    "__title__",
    "__version__",
    # Submodules
    "endpoints",
    "fastapi",
    "resources",
    "types",
    "utils",
]



from typing import AsyncIterator, Final, Optional

from typing_extensions import Unpack

from ncae_sdk._async.endpoint import AsyncEndpoint
from ncae_sdk._resource import ResourceId
from ncae_sdk._util import map_async
from ncae_sdk.resources._schema import (
    Report,
    ReportCreate,
    ReportCreateModel,
    ReportFilter,
    ReportFilterModel,
    ReportUpdate,
    ReportUpdateModel,
)


class AsyncReportEndpoint(AsyncEndpoint):
    BASE_PATH: Final[str] = "dashboard/v1/onetimereport"

    async def list(self, **filters: Unpack[ReportFilter]) -> AsyncIterator[Report]:
        results = self._list(self.BASE_PATH, ReportFilterModel, filters)
        return map_async(Report.parse_api, results)

    async def find(self, **filters: Unpack[ReportFilter]) -> Optional[Report]:
        result = await self._get_by_filters(self.BASE_PATH, ReportFilterModel, filters)
        return Report.parse_api(result) if result else None

    async def get(self, rid: ResourceId) -> Optional[Report]:
        result = await self._get_by_id(self.BASE_PATH, rid)
        return Report.parse_api(result) if result else None

    async def create(self, **payload: Unpack[ReportCreate]) -> Report:
        result = await self._create(self.BASE_PATH, ReportCreateModel, payload)
        return Report.parse_api(result)

    async def update(self, rid: ResourceId, /, **payload: Unpack[ReportUpdate]) -> Report:
        result = await self._update(self.BASE_PATH, rid, ReportUpdateModel, payload)
        return Report.parse_api(result)

    async def delete(self, rid: ResourceId) -> bool:
        return await self._delete(self.BASE_PATH, rid)



from collections.abc import AsyncIterator
from typing import Final, Optional

from typing_extensions import Unpack

from ncae_sdk._async.endpoint import AsyncEndpoint
from ncae_sdk._resource import ResourceId
from ncae_sdk._util import map_async
from ncae_sdk.resources._schema import (
    InfluxSystem,
    InfluxSystemCreate,
    InfluxSystemCreateModel,
    InfluxSystemFilter,
    InfluxSystemFilterModel,
    InfluxSystemUpdate,
    InfluxSystemUpdateModel,
)


class AsyncInfluxSystemEndpoint(AsyncEndpoint):
    BASE_PATH: Final[str] = "observability/v1/influx-system"

    async def list(self, **filters: Unpack[InfluxSystemFilter]) -> AsyncIterator[InfluxSystem]:
        results = self._list(self.BASE_PATH, InfluxSystemFilterModel, filters)
        return map_async(InfluxSystem.parse_api, results)

    async def find(self, **filters: Unpack[InfluxSystemFilter]) -> Optional[InfluxSystem]:
        result = await self._get_by_filters(self.BASE_PATH, InfluxSystemFilterModel, filters)
        return InfluxSystem.parse_api(result) if result else None

    async def get(self, rid: ResourceId) -> Optional[InfluxSystem]:
        result = await self._get_by_id(self.BASE_PATH, rid)
        return InfluxSystem.parse_api(result) if result else None

    async def create(self, **payload: Unpack[InfluxSystemCreate]) -> InfluxSystem:
        result = await self._create(self.BASE_PATH, InfluxSystemCreateModel, payload)
        return InfluxSystem.parse_api(result)

    async def update(self, rid: ResourceId, /, **payload: Unpack[InfluxSystemUpdate]) -> InfluxSystem:
        result = await self._update(self.BASE_PATH, rid, InfluxSystemUpdateModel, payload)
        return InfluxSystem.parse_api(result)

    async def delete(self, rid: ResourceId) -> bool:
        return await self._delete(self.BASE_PATH, rid)



import logging
from contextlib import asynccontextmanager
from ssl import SSLContext, VerifyMode
from typing import Any, AsyncIterator, Final, Optional

from httpcore import AsyncConnectionPool
from httpx import AsyncBaseTransport, AsyncClient
from typing_extensions import Self, Unpack

from ncae_sdk import __version__
from ncae_sdk._auth import Auth
from ncae_sdk._error import NcaeHttpError
from ncae_sdk._http import RequestArgs, Response
from ncae_sdk._session import SessionContext

logger = logging.getLogger("ncae_sdk.session")


class AsyncSession:
    __slots__ = ("_auth", "_client", "_context", "_is_clone", "_can_retry")

    def __init__(
        self,
        *,
        auth: Auth,
        context: SessionContext,
        client: AsyncClient,
        _is_clone: bool = False,
        _can_retry: bool = True,
    ) -> None:
        self._auth: Final[Auth] = auth
        self._client: Final[AsyncClient] = client
        self._context: Final[SessionContext] = context or SessionContext()
        self._is_clone: Final[bool] = _is_clone
        self._can_retry: Final[bool] = _can_retry

    @classmethod
    def create(
        cls,
        auth: Auth,
        base_url: str,
        timeout: int,
        verify: bool,
        httpx_transport: Optional[AsyncBaseTransport] = None,
    ) -> Self:
        return cls(
            auth=auth,
            context=SessionContext(),
            client=AsyncClient(
                base_url=cls._sanitize_base_url(base_url),
                headers={"User-Agent": f"ncae-sdk-python/{__version__}"},
                timeout=timeout,
                verify=verify,
                transport=httpx_transport,
            ),
        )

    @property
    def context(self) -> SessionContext:
        return self._context

    @property
    def base_url(self) -> str:
        return str(self._client.base_url)

    @property
    def auth(self) -> Auth:
        return self._auth

    @property
    def timeout(self) -> float:
        return self._client.timeout.connect or 0

    @property
    def verify(self) -> bool:
        httpx_pool = getattr(self._client._transport, "_pool", None)
        assert isinstance(httpx_pool, AsyncConnectionPool)
        assert isinstance(httpx_pool._ssl_context, SSLContext)
        return httpx_pool._ssl_context.verify_mode != VerifyMode.CERT_NONE

    def clone(self, context: Optional[SessionContext] = None, *, _can_retry: Optional[bool] = None) -> Self:
        return self.__class__(
            context=context if context is not None else self._context,
            client=self._client,
            auth=self._auth,
            _is_clone=True,
            _can_retry=_can_retry if _can_retry is not None else self._can_retry,
        )

    async def close(self) -> None:
        if not self._is_clone:
            await self._client.aclose()

    async def get(self, url: str, **kwargs: Unpack[RequestArgs]) -> Response:
        return await self.request("GET", url, **kwargs)

    async def post(self, url: str, **kwargs: Unpack[RequestArgs]) -> Response:
        return await self.request("POST", url, **kwargs)

    async def patch(self, url: str, **kwargs: Unpack[RequestArgs]) -> Response:
        return await self.request("PATCH", url, **kwargs)

    async def delete(self, url: str, **kwargs: Unpack[RequestArgs]) -> Response:
        return await self.request("DELETE", url, **kwargs)

    async def request(self, method: str, url: str, **kwargs: Unpack[RequestArgs]) -> Response:
        logger.debug("Executing NCAE API request: %s %s", method, url)
        mapped_kwargs = self._map_httpx_kwargs(kwargs)
        native_response = await self._client.request(method=method, url=url, **mapped_kwargs)
        response = Response(native_response)

        if response.status_code == 401 and self._can_retry:
            retry_session = self.clone(_can_retry=False)
            await self._auth.async_authenticate(retry_session)
            return await retry_session.request(method=method, url=url, **kwargs)

        if not response.is_success:
            raise NcaeHttpError(response)

        return response

    @asynccontextmanager
    async def stream(self, method: str, url: str, **kwargs: Unpack[RequestArgs]) -> AsyncIterator[Response]:
        mapped_kwargs = self._map_httpx_kwargs(kwargs)
        for attempt in range(2):
            async with self._client.stream(method=method, url=url, **mapped_kwargs) as native_response:
                response = Response(native_response)
                if response.status_code == 401:
                    await self._auth.async_authenticate(self)
                    continue

                # If the response has an unexpected status code, read the whole body, then raise a regular error.
                # It's important to read the body at this point, otherwise NcaeHttpError is missing information.
                if not response.is_success:
                    await response.aread()
                    raise NcaeHttpError(response)

                # Break the loop after the first successful response.
                yield response
                break

    def _map_httpx_kwargs(self, kwargs: RequestArgs) -> dict[str, Any]:
        mapped_kwargs: dict[str, Any] = dict(kwargs.copy())
        mapped_kwargs["params"] = mapped_kwargs.pop("query", None)

        if self._context.request_headers:
            headers = mapped_kwargs.get("headers", {}).copy()
            headers.update(self._context.request_headers)
            mapped_kwargs["headers"] = headers

        return mapped_kwargs

    @classmethod
    def _sanitize_base_url(cls, base_url: str) -> str:
        base_url = base_url.strip().rstrip("/") + "/"
        if not base_url.endswith("/api/"):
            base_url += "api/"

        return base_url



from collections.abc import AsyncIterator
from typing import Final, Optional

from typing_extensions import Unpack, deprecated

from ncae_sdk._async.endpoint import AsyncEndpoint
from ncae_sdk._resource import ResourceId
from ncae_sdk._util import map_async
from ncae_sdk.resources._schema import (
    Phase,
    PhaseCreate,
    PhaseCreateModel,
    PhaseFilter,
    PhaseFilterModel,
    PhaseUpdate,
    PhaseUpdateModel,
)


@deprecated(
    "This endpoint is deprecated and should not be used in new code. While there is no set date for removal, "
    "it is recommended to switch to the alternative endpoint. The alternative endpoint is: /automation/phase-type/"
)
class AsyncPhaseEndpoint(AsyncEndpoint):
    BASE_PATH: Final[str] = "automation/v1/phase"

    async def list(self, **filters: Unpack[PhaseFilter]) -> AsyncIterator[Phase]:
        results = self._list(self.BASE_PATH, PhaseFilterModel, filters)
        return map_async(Phase.parse_api, results)

    async def find(self, **filters: Unpack[PhaseFilter]) -> Optional[Phase]:
        result = await self._get_by_filters(self.BASE_PATH, PhaseFilterModel, filters)
        return Phase.parse_api(result) if result else None

    async def get(self, rid: ResourceId) -> Optional[Phase]:
        result = await self._get_by_id(self.BASE_PATH, rid)
        return Phase.parse_api(result) if result else None

    async def create(self, **payload: Unpack[PhaseCreate]) -> Phase:
        result = await self._create(self.BASE_PATH, PhaseCreateModel, payload)
        return Phase.parse_api(result)

    async def update(self, rid: ResourceId, /, **payload: Unpack[PhaseUpdate]) -> Phase:
        result = await self._update(self.BASE_PATH, rid, PhaseUpdateModel, payload)
        return Phase.parse_api(result)

    async def delete(self, rid: ResourceId) -> bool:
        return await self._delete(self.BASE_PATH, rid)



import time
from typing import Awaitable, Callable
from uuid import uuid4

from fastapi import Request, Response

from ncae_sdk.fastapi._utils import obtain_request_logger

CallNext = Callable[[Request], Awaitable[Response]]


async def add_tracing_headers(request: Request, call_next: CallNext) -> Response:
    correlation_id = request.headers.get("X-Correlation-Id") or str(uuid4())
    request.state.correlation_id = correlation_id

    start_time = time.perf_counter()
    response = await call_next(request)
    response_time = time.perf_counter() - start_time

    response.headers["X-Correlation-Id"] = correlation_id
    response.headers["Server-Timing"] = f"request;dur={response_time:.3f}"

    return response


async def inject_request_logger(request: Request, call_next: CallNext) -> Response:
    obtain_request_logger(request, recreate=True)
    return await call_next(request)


async def log_http_request(request: Request, call_next: CallNext) -> Response:
    response = await call_next(request)
    if request.method in ["HEAD", "GET"] and request.url.path == "/api/system/healthz":
        return response

    logger = obtain_request_logger(request)
    logger.info(
        "%d %s %s from %s",
        response.status_code,
        request.method,
        request.url.path,
        request.client.host if request.client else "unknown",
    )

    return response



from fastapi import APIRouter, FastAPI, Request
from starlette.responses import RedirectResponse

system_router = APIRouter(prefix="/api/system", tags=["system"])


@system_router.get("/", include_in_schema=False)
async def root() -> RedirectResponse:
    return RedirectResponse(url="/api/docs")


@system_router.get("/healthz")
async def health_check(request: Request) -> dict[str, str]:
    app = request.app
    assert isinstance(app, FastAPI)

    return {
        "status": "ok",
        "title": getattr(app, "title"),
        "version": getattr(app, "version"),
    }



# @generated
# This file was automatically generated by the ncae-sdk code generator.
# DO NOT EDIT MANUALLY, ANY CHANGES WILL BE OVERWRITTEN.

from collections.abc import Iterator
from typing import Final, Optional

from typing_extensions import Unpack

from ncae_sdk._resource import ResourceId
from ncae_sdk._sync.endpoint import Endpoint
from ncae_sdk._util import map_sync
from ncae_sdk.resources._schema import (
    DeviceModel,
    DeviceModelFilter,
    DeviceModelFilterModel,
)


class DeviceModelEndpoint(Endpoint):
    BASE_PATH: Final[str] = "dashboard/v1/device-model"

    def list(self, **filters: Unpack[DeviceModelFilter]) -> Iterator[DeviceModel]:
        results = self._list(self.BASE_PATH, DeviceModelFilterModel, filters)
        return map_sync(DeviceModel.parse_api, results)

    def find(self, **filters: Unpack[DeviceModelFilter]) -> Optional[DeviceModel]:
        result = self._get_by_filters(self.BASE_PATH, DeviceModelFilterModel, filters)
        return DeviceModel.parse_api(result) if result else None

    def get(self, rid: ResourceId) -> Optional[DeviceModel]:
        result = self._get_by_id(self.BASE_PATH, rid)
        return DeviceModel.parse_api(result) if result else None



import inspect
from typing import Any, AsyncIterator, Callable, Iterator, Optional, TypeVar

T = TypeVar("T")
R = TypeVar("R")


map_sync = map


def ensure_callable_sync(func: Callable[..., Any]) -> None:
    if not callable(func):
        raise TypeError(f"Function {func.__name__} must be a callable (expected [def], got [{type(func).__name__}]).")
    if inspect.iscoroutinefunction(func):
        raise TypeError(f"Function {func.__name__} must be a regular function (expected [def], got [async def]).")


def ensure_callable_async(func: Callable[..., Any]) -> None:
    if not callable(func):
        raise TypeError(
            f"Function {func.__name__} must be a callable (expected [async def], got [{type(func).__name__}])."
        )
    if not inspect.iscoroutinefunction(func):
        raise TypeError(f"Function {func.__name__} must be a coroutine function (expected [async def], got [def]).")


def parse_env_bool(value: str) -> bool:
    value = value.lower().strip()
    if value in ("true", "1", "yes", "on"):
        return True
    elif value in ("false", "0", "no", "off"):
        return False
    else:
        raise ValueError(f"Invalid boolean value: {value}")


def parse_env_int(value: str) -> int:
    try:
        return int(value.strip())
    except ValueError as e:
        raise ValueError(f"Invalid integer value: {value}") from e


async def map_async(func: Callable[[T], R], iterable: AsyncIterator[T]) -> AsyncIterator[R]:
    async for item in iterable:
        yield await func(item) if inspect.iscoroutinefunction(func) else func(item)


async def next_async(iterable: AsyncIterator[T], default: Optional[T] = None) -> Optional[T]:
    try:
        return await iterable.__anext__()
    except StopAsyncIteration:
        return None


def next_sync(iterable: Iterator[T], default: Optional[T] = None) -> Optional[T]:
    return next(iterable, default)


__all__ = [
    "parse_env_bool",
    "parse_env_int",
    "ensure_callable_sync",
    "map_sync",
    "next_sync",
    "ensure_callable_async",
    "map_async",
    "next_async",
]



from typing import TYPE_CHECKING, Final

if TYPE_CHECKING:
    from ncae_sdk._http import Response


class NcaeHttpError(Exception):
    def __init__(self, response: "Response") -> None:
        self.url: Final[str] = str(response.request.url)
        self.status_code: Final[int] = response.status_code
        self.content: Final[str] = response.content.decode("utf-8", errors="replace")
        super().__init__(
            f"Request to [{self.url}] failed with status code {self.status_code} and content: {self.content}"
        )



# @generated
# This file was automatically generated by the ncae-sdk code generator.
# DO NOT EDIT MANUALLY, ANY CHANGES WILL BE OVERWRITTEN.

import logging
from contextlib import contextmanager
from ssl import SSLContext, VerifyMode
from typing import Any, Final, Iterator, Optional

from httpcore import ConnectionPool
from httpx import BaseTransport, Client
from typing_extensions import Self, Unpack

from ncae_sdk import __version__
from ncae_sdk._auth import Auth
from ncae_sdk._error import NcaeHttpError
from ncae_sdk._http import RequestArgs, Response
from ncae_sdk._session import SessionContext

logger = logging.getLogger("ncae_sdk.session")


class Session:
    __slots__ = ("_auth", "_client", "_context", "_is_clone", "_can_retry")

    def __init__(
        self,
        *,
        auth: Auth,
        context: SessionContext,
        client: Client,
        _is_clone: bool = False,
        _can_retry: bool = True,
    ) -> None:
        self._auth: Final[Auth] = auth
        self._client: Final[Client] = client
        self._context: Final[SessionContext] = context or SessionContext()
        self._is_clone: Final[bool] = _is_clone
        self._can_retry: Final[bool] = _can_retry

    @classmethod
    def create(
        cls,
        auth: Auth,
        base_url: str,
        timeout: int,
        verify: bool,
        httpx_transport: Optional[BaseTransport] = None,
    ) -> Self:
        return cls(
            auth=auth,
            context=SessionContext(),
            client=Client(
                base_url=cls._sanitize_base_url(base_url),
                headers={"User-Agent": f"ncae-sdk-python/{__version__}"},
                timeout=timeout,
                verify=verify,
                transport=httpx_transport,
            ),
        )

    @property
    def context(self) -> SessionContext:
        return self._context

    @property
    def base_url(self) -> str:
        return str(self._client.base_url)

    @property
    def auth(self) -> Auth:
        return self._auth

    @property
    def timeout(self) -> float:
        return self._client.timeout.connect or 0

    @property
    def verify(self) -> bool:
        httpx_pool = getattr(self._client._transport, "_pool", None)
        assert isinstance(httpx_pool, ConnectionPool)
        assert isinstance(httpx_pool._ssl_context, SSLContext)
        return httpx_pool._ssl_context.verify_mode != VerifyMode.CERT_NONE

    def clone(self, context: Optional[SessionContext] = None, *, _can_retry: Optional[bool] = None) -> Self:
        return self.__class__(
            context=context if context is not None else self._context,
            client=self._client,
            auth=self._auth,
            _is_clone=True,
            _can_retry=_can_retry if _can_retry is not None else self._can_retry,
        )

    def close(self) -> None:
        if not self._is_clone:
            self._client.close()

    def get(self, url: str, **kwargs: Unpack[RequestArgs]) -> Response:
        return self.request("GET", url, **kwargs)

    def post(self, url: str, **kwargs: Unpack[RequestArgs]) -> Response:
        return self.request("POST", url, **kwargs)

    def patch(self, url: str, **kwargs: Unpack[RequestArgs]) -> Response:
        return self.request("PATCH", url, **kwargs)

    def delete(self, url: str, **kwargs: Unpack[RequestArgs]) -> Response:
        return self.request("DELETE", url, **kwargs)

    def request(self, method: str, url: str, **kwargs: Unpack[RequestArgs]) -> Response:
        logger.debug("Executing NCAE API request: %s %s", method, url)
        mapped_kwargs = self._map_httpx_kwargs(kwargs)
        native_response = self._client.request(method=method, url=url, **mapped_kwargs)
        response = Response(native_response)

        if response.status_code == 401 and self._can_retry:
            retry_session = self.clone(_can_retry=False)
            self._auth.authenticate(retry_session)
            return retry_session.request(method=method, url=url, **kwargs)

        if not response.is_success:
            raise NcaeHttpError(response)

        return response

    @contextmanager
    def stream(self, method: str, url: str, **kwargs: Unpack[RequestArgs]) -> Iterator[Response]:
        mapped_kwargs = self._map_httpx_kwargs(kwargs)
        for attempt in range(2):
            with self._client.stream(method=method, url=url, **mapped_kwargs) as native_response:
                response = Response(native_response)
                if response.status_code == 401:
                    self._auth.authenticate(self)
                    continue

                # If the response has an unexpected status code, read the whole body, then raise a regular error.
                # It's important to read the body at this point, otherwise NcaeHttpError is missing information.
                if not response.is_success:
                    response.read()
                    raise NcaeHttpError(response)

                # Break the loop after the first successful response.
                yield response
                break

    def _map_httpx_kwargs(self, kwargs: RequestArgs) -> dict[str, Any]:
        mapped_kwargs: dict[str, Any] = dict(kwargs.copy())
        mapped_kwargs["params"] = mapped_kwargs.pop("query", None)

        if self._context.request_headers:
            headers = mapped_kwargs.get("headers", {}).copy()
            headers.update(self._context.request_headers)
            mapped_kwargs["headers"] = headers

        return mapped_kwargs

    @classmethod
    def _sanitize_base_url(cls, base_url: str) -> str:
        base_url = base_url.strip().rstrip("/") + "/"
        if not base_url.endswith("/api/"):
            base_url += "api/"

        return base_url



from typing import Final, Optional

from fastapi import Request
from typing_extensions import Generic, Self

from ncae_sdk._session import SessionContext
from ncae_sdk.fastapi._context import BaseContext
from ncae_sdk.fastapi._dependencies import AsyncNcaeClientDep, DependencyMixin, NcaeClientDep, ThreadPoolExecutorDep
from ncae_sdk.fastapi.extapi._context_async import AsyncExtApiPhaseContextMixin
from ncae_sdk.fastapi.extapi._context_sync import ExtApiPhaseContextMixin
from ncae_sdk.fastapi.extapi._contracts import (
    CallbackT,
    CmdbT,
    ExtApiDevice,
    ExtApiPhaseExtraVars,
    ExtApiPhaseRequest,
    ExtApiReportRequest,
    QueryT,
)
from ncae_sdk.fastapi.extapi._protocols import ExtApiPhaseContextProtocol
from ncae_sdk.fastapi.runner._parallel import ParallelRunner
from ncae_sdk.fastapi.runner._utils import ItemT


class ExtApiPhaseContext(
    ExtApiPhaseContextMixin[CmdbT, CallbackT],
    AsyncExtApiPhaseContextMixin[CmdbT, CallbackT],
    DependencyMixin,
    BaseContext,
    ExtApiPhaseContextProtocol[CmdbT, CallbackT],
    Generic[CmdbT, CallbackT],
):
    def __init__(
        self,
        body: ExtApiPhaseRequest[CmdbT, CallbackT],
        request: Request,
        ncae_client: NcaeClientDep,
        async_ncae_client: AsyncNcaeClientDep,
        thread_pool: ThreadPoolExecutorDep,
    ) -> None:
        self._body: Final[ExtApiPhaseRequest[CmdbT, CallbackT]] = body

        super().__init__(
            ncae_client=ncae_client,
            async_ncae_client=async_ncae_client,
            request=request,
            thread_pool=thread_pool,
        )

    @property
    def is_decommission(self) -> bool:
        return self._body.extra_vars.is_decommission

    @property
    def cmdb_data(self) -> CmdbT:
        return self._body.extra_vars.cmdb_data

    @property
    def cmdb_data_previous(self) -> Optional[CmdbT]:
        return self._body.extra_vars.cmdb_data_previous

    @property
    def callback_data(self) -> CallbackT:
        return self._body.extra_vars.callback_data

    @property
    def devices(self) -> list[ExtApiDevice]:
        return self._body.extra_vars.devices

    @property
    def body(self) -> ExtApiPhaseExtraVars[CmdbT, CallbackT]:
        return self._body.extra_vars

    def get_device_runner(self) -> ParallelRunner[Self, ExtApiDevice]:
        return self.get_parallel_runner(self._body.extra_vars.devices)

    def get_parallel_runner(self, items: list[ItemT]) -> ParallelRunner[Self, ItemT]:
        return ParallelRunner(
            ctx=self,
            items=items,
            hook_sync=self._publish_progress_update,
            hook_async=self._async_publish_progress_update,
        )

    def _build_session_context(self) -> SessionContext:
        return SessionContext(
            ncae_tenant_id=self._body.extra_vars.ncae_tenant_id,
            ncae_transaction_id=self._body.extra_vars.ncae_transaction_id,
        )


class ExtApiReportContext(DependencyMixin, BaseContext, Generic[QueryT]):
    def __init__(
        self,
        body: ExtApiReportRequest[QueryT],
        request: Request,
        ncae_client: NcaeClientDep,
        async_ncae_client: AsyncNcaeClientDep,
        thread_pool: ThreadPoolExecutorDep,
    ) -> None:
        self._body: Final[ExtApiReportRequest[QueryT]] = body

        super().__init__(
            ncae_client=ncae_client,
            async_ncae_client=async_ncae_client,
            request=request,
            thread_pool=thread_pool,
        )

    @property
    def body(self) -> ExtApiReportRequest[QueryT]:
        return self._body

    @property
    def devices(self) -> list[ExtApiDevice]:
        return self._body.devices

    @property
    def query(self) -> QueryT:
        return self._body.query



from dataclasses import dataclass
from typing import Generic, Optional, TypeVar, cast

from ncae_sdk.fastapi._context import BaseContext

T = TypeVar("T")
ItemT = TypeVar("ItemT")
ContextT = TypeVar("ContextT", bound=BaseContext)


@dataclass(frozen=True)
class RunnerResult(Generic[ItemT, T]):
    item: ItemT
    exception: Optional[Exception]
    raw_value: Optional[T]

    @property
    def ok(self) -> bool:
        return self.exception is None

    @property
    def value(self) -> T:
        if self.exception is not None:
            raise self.exception

        return cast(T, self.raw_value)



from __future__ import annotations

from typing_extensions import TypeAlias

from ncae_sdk.utils.jinja2_tplgen._ast_collect import IterablePath, VariablePath, collect_variables
from ncae_sdk.utils.jinja2_tplgen._service_models import GroupTemplate, Template, TextListTemplate, TextTemplate

# Recursive mapping representing a variable namespace tree.
NodeTree: TypeAlias = dict[str, "NodeTree"]


def to_service_template(jinja2_template: str) -> list[Template]:
    """
    Converts the specified Jinja2 template, specified as a string, into a NCAE service template.
    All variables are turned into appropriate fields to match the structure of the Jinja2 template.
    """
    paths, iterables = collect_variables(jinja2_template)
    path_list = sorted(paths)
    iterable_list = sorted(iterables)
    tree = _build_node_tree(path_list)

    return _to_templates(tree, iterable_list)


def _build_node_tree(paths: list[VariablePath]) -> NodeTree:
    """
    Expands the flat tuple paths from the variable collector into a nested dict tree structure.
    Example input: [('user', 'name'), ('user', 'email'), ('settings', 'tz')]
    Example output: {'user': {'name': {}, 'email': {}}, 'settings': {'tz': {}}}
    """
    tree: NodeTree = {}
    for path in paths:
        current: NodeTree = tree
        for part in path:
            current = current.setdefault(part, {})

    return tree


def _to_templates(tree: NodeTree, iterables: list[IterablePath]) -> list[Template]:
    """
    Transforms a node tree into a list of NCAE service template fields with appropriate types.
    """
    iterable_set = set(iterables)

    def node_to_template(name: str, subtree: NodeTree, path: VariablePath) -> Template:
        label = _humanize_name_to_label(name)

        # If this node has its own subtree, treat it as a group input with children
        # This applies both to lists and dicts, in the latter case a separate text input is created for each leaf node
        if subtree:
            children = [
                node_to_template(child_name, child_tree, path + (child_name,))
                for child_name, child_tree in sorted(subtree.items())
            ]
            return GroupTemplate(type="group", name=name, label=label, children=children)

        # If this node is itself part of an iterable, then the input must be a text-list to allow multiple values
        # E.g. a top-level "{% for p in port %}{{ port }}{% endfor %}"
        if path in iterable_set:
            return TextListTemplate(type="text-list", name=name, label=label)

        # In all other cases, default to a text input
        return TextTemplate(type="text", name=name, label=label)

    return [node_to_template(root_name, subtree, (root_name,)) for root_name, subtree in sorted(tree.items())]


def _humanize_name_to_label(name: str) -> str:
    """
    Tries to convert a snake_case or kebab-case name into a human-friendly label.
    E.g., "interface_name" -> "Interface name"
    """
    s = name.replace("_", " ").replace("-", " ")
    return s[:1].upper() + s[1:]



# @generated
# This file was automatically generated by the ncae-sdk code generator.
# DO NOT EDIT MANUALLY, ANY CHANGES WILL BE OVERWRITTEN.

import os
from functools import cached_property
from types import TracebackType
from typing import Optional, Type

from httpx import BaseTransport
from typing_extensions import Self

from ncae_sdk._auth import Auth, SessionAuth
from ncae_sdk._session import SessionContext
from ncae_sdk._sync.session import Session
from ncae_sdk._util import parse_env_bool, parse_env_int
from ncae_sdk.endpoints import (
    AuditLogEndpoint,
    AwxModuleEndpoint,
    BasePhaseEndpoint,
    ChangeScheduleEndpoint,
    ChangeScheduleRunEndpoint,
    CmdbEntryEndpoint,
    CredentialEndpoint,
    DeviceEndpoint,
    DeviceGroupEndpoint,
    DeviceModelEndpoint,
    ExtApiAuthEndpoint,
    ExtApiPhaseEndpoint,
    ExtApiServiceEndpoint,
    GrafanaDashboardEndpoint,
    GrafanaSystemEndpoint,
    InfluxRepositoryEndpoint,
    InfluxSystemEndpoint,
    ModuleEndpoint,
    PhaseEndpoint,
    PhaseInstanceEndpoint,
    PollingPhaseEndpoint,
    PythonScriptPhaseEndpoint,
    ServiceEndpoint,
    ServiceInstanceEndpoint,
    ServiceInstanceLogEndpoint,
    ServiceManagementPhaseEndpoint,
    SharedDataBucketEndpoint,
    SharedDataSubscriberEndpoint,
    TagEndpoint,
    TagRelationEndpoint,
    TenantEndpoint,
    TransactionEndpoint,
    UploadFileEndpoint,
    UsageLogEndpoint,
    UserEndpoint,
)
from ncae_sdk.endpoints._sync.favourite import FavouriteEndpoint
from ncae_sdk.endpoints._sync.report import ReportEndpoint


class Client:
    def __init__(
        self,
        base_url: Optional[str] = None,
        auth: Optional[Auth] = None,
        timeout: Optional[int] = None,
        verify: Optional[bool] = None,
        httpx_transport: Optional[BaseTransport] = None,
    ) -> None:
        """
        Initializes a new asynchronous client for interacting with the NCAE API.

        This client provides access to various endpoints such as audit logs, devices, credentials, and more. It is
        based on an underlying session that uses httpx for asynchronous HTTP requests. For these reasons, it is
        recommended to use this client as a context manager, as this ensures cleanup and allows proper pooling.

        All arguments are marked as optional, which will fallback to environment variables or default values.
        If no suitable configuration is found, an error will be raised during instantiation.

        :param base_url: The base URL of the NCAE API, e.g. `https://ncae.example.com/api/`.
        :param auth: An instance of :class:`Auth` that provides authentication for the API requests.
        :param timeout: The desired timeout for API requests in seconds. Usage of default value is recommended.
        :param verify: Whether to verify SSL certificates. Set to `False` if you are using a self-signed certificate.
        """
        if not base_url:
            base_url = os.environ.get("NCAE_SDK_BASE_URL")
            if not base_url:
                raise ValueError("Unable to determine NCAE SDK base URL.")

        if not auth:
            username = os.environ.get("NCAE_SDK_USERNAME")
            password = os.environ.get("NCAE_SDK_PASSWORD")
            if not username or not password:
                raise ValueError("Unable to determine NCAE SDK authentication credentials.")

            auth = SessionAuth(username=username, password=password)

        if timeout is None:
            if value := os.environ.get("NCAE_SDK_HTTP_TIMEOUT"):
                timeout = parse_env_int(value)
            else:
                timeout = 5

        if verify is None:
            if value := os.environ.get("NCAE_SDK_HTTP_VERIFY"):
                verify = parse_env_bool(value)
            else:
                verify = True

        self._session: Session = Session.create(
            base_url=base_url,
            auth=auth,
            timeout=timeout,
            verify=verify,
            httpx_transport=httpx_transport,
        )

    def clone_with_context(self, context: SessionContext) -> Self:
        instance = self.__class__.__new__(self.__class__)
        instance._session = self._session.clone(context)
        return instance

    def close(self) -> None:
        self._session.close()

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.close()

    def ping(self) -> None:
        self._session.get("readyz")

    @property
    def session(self) -> Session:
        return self._session

    @cached_property
    def audit_logs(self) -> AuditLogEndpoint:
        return AuditLogEndpoint(session=self._session)

    @cached_property
    def awx_modules(self) -> AwxModuleEndpoint:
        return AwxModuleEndpoint(session=self._session)

    @cached_property
    def base_phases(self) -> BasePhaseEndpoint:
        return BasePhaseEndpoint(session=self._session)

    @cached_property
    def change_schedules(self) -> ChangeScheduleEndpoint:
        return ChangeScheduleEndpoint(session=self._session)

    @cached_property
    def change_schedule_runs(self) -> ChangeScheduleRunEndpoint:
        return ChangeScheduleRunEndpoint(session=self._session)

    @cached_property
    def cmdb_entries(self) -> CmdbEntryEndpoint:
        return CmdbEntryEndpoint(session=self._session)

    @cached_property
    def credentials(self) -> CredentialEndpoint:
        return CredentialEndpoint(session=self._session)

    @cached_property
    def devices(self) -> DeviceEndpoint:
        return DeviceEndpoint(session=self._session)

    @cached_property
    def device_groups(self) -> DeviceGroupEndpoint:
        return DeviceGroupEndpoint(session=self._session)

    @cached_property
    def device_models(self) -> DeviceModelEndpoint:
        return DeviceModelEndpoint(session=self._session)

    @cached_property
    def ext_api_auths(self) -> ExtApiAuthEndpoint:
        return ExtApiAuthEndpoint(session=self._session)

    @cached_property
    def ext_api_services(self) -> ExtApiServiceEndpoint:
        return ExtApiServiceEndpoint(session=self._session)

    @cached_property
    def ext_api_phases(self) -> ExtApiPhaseEndpoint:
        return ExtApiPhaseEndpoint(session=self._session)

    @cached_property
    def favourites(self) -> FavouriteEndpoint:
        return FavouriteEndpoint(session=self._session)

    @cached_property
    def grafana_dashboards(self) -> GrafanaDashboardEndpoint:
        return GrafanaDashboardEndpoint(session=self._session)

    @cached_property
    def grafana_systems(self) -> GrafanaSystemEndpoint:  # noqa: F821
        return GrafanaSystemEndpoint(session=self._session)

    @cached_property
    def influx_repositories(self) -> InfluxRepositoryEndpoint:
        return InfluxRepositoryEndpoint(session=self._session)

    @cached_property
    def influx_systems(self) -> InfluxSystemEndpoint:
        return InfluxSystemEndpoint(session=self._session)

    @cached_property
    def modules(self) -> ModuleEndpoint:
        return ModuleEndpoint(session=self._session)

    @cached_property
    def phases(self) -> PhaseEndpoint:
        return PhaseEndpoint(session=self._session)

    @cached_property
    def phase_instances(self) -> PhaseInstanceEndpoint:
        return PhaseInstanceEndpoint(session=self._session)

    @cached_property
    def polling_phases(self) -> PollingPhaseEndpoint:
        return PollingPhaseEndpoint(session=self._session)

    @cached_property
    def python_script_phases(self) -> PythonScriptPhaseEndpoint:
        return PythonScriptPhaseEndpoint(session=self._session)

    @cached_property
    def reports(self) -> ReportEndpoint:
        return ReportEndpoint(session=self._session)

    @cached_property
    def services(self) -> ServiceEndpoint:
        return ServiceEndpoint(session=self._session)

    @cached_property
    def service_instances(self) -> ServiceInstanceEndpoint:
        return ServiceInstanceEndpoint(session=self._session)

    @cached_property
    def service_instance_logs(self) -> ServiceInstanceLogEndpoint:
        return ServiceInstanceLogEndpoint(session=self._session)

    @cached_property
    def shared_data_buckets(self) -> SharedDataBucketEndpoint:
        return SharedDataBucketEndpoint(session=self._session)

    @cached_property
    def shared_data_subscribers(self) -> SharedDataSubscriberEndpoint:
        return SharedDataSubscriberEndpoint(session=self._session)

    @cached_property
    def service_management_phases(self) -> ServiceManagementPhaseEndpoint:
        return ServiceManagementPhaseEndpoint(session=self._session)

    @cached_property
    def tags(self) -> TagEndpoint:
        return TagEndpoint(session=self._session)

    @cached_property
    def tag_relations(self) -> TagRelationEndpoint:
        return TagRelationEndpoint(session=self._session)

    @cached_property
    def tenants(self) -> TenantEndpoint:
        return TenantEndpoint(session=self._session)

    @cached_property
    def transactions(self) -> TransactionEndpoint:
        return TransactionEndpoint(session=self._session)

    @cached_property
    def upload_files(self) -> UploadFileEndpoint:
        return UploadFileEndpoint(session=self._session)

    @cached_property
    def usage_logs(self) -> UsageLogEndpoint:
        return UsageLogEndpoint(session=self._session)

    @cached_property
    def users(self) -> UserEndpoint:
        return UserEndpoint(session=self._session)



from collections.abc import AsyncIterator
from typing import Final, Optional

from typing_extensions import Unpack

from ncae_sdk._async.endpoint import AsyncEndpoint
from ncae_sdk._resource import ResourceId
from ncae_sdk._util import map_async
from ncae_sdk.resources._schema import Module, ModuleFilter, ModuleFilterModel


class AsyncModuleEndpoint(AsyncEndpoint):
    BASE_PATH: Final[str] = "admin/v1/module"

    async def list(self, **filters: Unpack[ModuleFilter]) -> AsyncIterator[Module]:
        results = self._list(self.BASE_PATH, ModuleFilterModel, filters)
        return map_async(Module.parse_api, results)

    async def find(self, **filters: Unpack[ModuleFilter]) -> Optional[Module]:
        result = await self._get_by_filters(self.BASE_PATH, ModuleFilterModel, filters)
        return Module.parse_api(result) if result else None

    async def get(self, rid: ResourceId) -> Optional[Module]:
        result = await self._get_by_id(self.BASE_PATH, rid)
        return Module.parse_api(result) if result else None



# @generated
# This file was automatically generated by the ncae-sdk code generator.
# DO NOT EDIT MANUALLY, ANY CHANGES WILL BE OVERWRITTEN.

from collections.abc import Iterator
from typing import Final, Optional

from typing_extensions import Unpack

from ncae_sdk._resource import ResourceId
from ncae_sdk._sync.endpoint import Endpoint
from ncae_sdk._util import map_sync
from ncae_sdk.resources._schema import (
    GrafanaSystem,
    GrafanaSystemCreate,
    GrafanaSystemCreateModel,
    GrafanaSystemFilter,
    GrafanaSystemFilterModel,
    GrafanaSystemUpdate,
    GrafanaSystemUpdateModel,
)


class GrafanaSystemEndpoint(Endpoint):
    BASE_PATH: Final[str] = "observability/v1/grafana-system"

    def list(self, **filters: Unpack[GrafanaSystemFilter]) -> Iterator[GrafanaSystem]:
        results = self._list(self.BASE_PATH, GrafanaSystemFilterModel, filters)
        return map_sync(GrafanaSystem.parse_api, results)

    def find(self, **filters: Unpack[GrafanaSystemFilter]) -> Optional[GrafanaSystem]:
        result = self._get_by_filters(self.BASE_PATH, GrafanaSystemFilterModel, filters)
        return GrafanaSystem.parse_api(result) if result else None

    def get(self, rid: ResourceId) -> Optional[GrafanaSystem]:
        result = self._get_by_id(self.BASE_PATH, rid)
        return GrafanaSystem.parse_api(result) if result else None

    def create(self, **payload: Unpack[GrafanaSystemCreate]) -> GrafanaSystem:
        result = self._create(self.BASE_PATH, GrafanaSystemCreateModel, payload)
        return GrafanaSystem.parse_api(result)

    def update(self, rid: ResourceId, /, **payload: Unpack[GrafanaSystemUpdate]) -> GrafanaSystem:
        result = self._update(self.BASE_PATH, rid, GrafanaSystemUpdateModel, payload)
        return GrafanaSystem.parse_api(result)

    def delete(self, rid: ResourceId) -> bool:
        return self._delete(self.BASE_PATH, rid)



from typing import IO, Any, AsyncIterator, Callable, Final, Mapping, Optional, Union

from pydantic import TypeAdapter
from typing_extensions import TypeAlias, Unpack

from ncae_sdk._async.session import AsyncSession
from ncae_sdk._error import NcaeHttpError
from ncae_sdk._http import RequestArgs, Response
from ncae_sdk._resource import ResourceId
from ncae_sdk._util import next_async

PreProcessor: TypeAlias = Callable[[dict[str, Any]], dict[str, Any]]
PostProcessor: TypeAlias = Callable[[Response], Any]


class AsyncEndpoint:
    __slots__ = ("_session",)

    DEFAULT_PAGE_SIZE: int = 250

    def __init__(self, *, session: AsyncSession) -> None:
        self._session: Final[AsyncSession] = session

    async def _refetch(self, url: str, response: Response) -> Any:
        if response.status_code == 204 or (response.status_code == 200 and len(response.content) == 0):
            return None

        response_json = response.json()
        if isinstance(response_json, dict) and "id" in response_json:
            return await self._get_by_id(url, response_json["id"])

        return response_json

    async def _create(
        self,
        url: str,
        create_model: TypeAdapter[Any],
        payload: Any,
        *,
        preprocessor: Optional[PreProcessor] = None,
    ) -> Any:
        payload_data = self._transform_data(create_model, payload)
        payload_data = preprocessor(payload_data) if preprocessor else payload_data
        response = await self._session.post(url, json=payload_data)
        return await self._refetch(url, response)

    async def _create_multipart(
        self,
        url: str,
        create_model: TypeAdapter[Any],
        payload: Any,
        files: Mapping[str, IO[bytes]],
        *,
        preprocessor: Optional[PreProcessor] = None,
    ) -> Any:
        payload_data = self._transform_data(create_model, payload)
        payload_data = preprocessor(payload_data) if preprocessor else payload_data
        response = await self._session.post(url, data=payload_data, files=files)
        return await self._refetch(url, response)

    async def _update(
        self,
        url: str,
        rid: ResourceId,
        update_model: TypeAdapter[Any],
        payload: Any,
        *,
        preprocessor: Optional[PreProcessor] = None,
    ) -> Any:
        payload_data = self._transform_data(update_model, payload)
        payload_data = preprocessor(payload_data) if preprocessor else payload_data
        response = await self._session.patch(f"{url}/{rid}", json=payload_data)
        return await self._refetch(url, response)

    async def _delete(self, url: str, rid: ResourceId) -> bool:
        try:
            await self._session.delete(f"{url}/{rid}")
            return True
        except NcaeHttpError as exc:
            if exc.status_code == 404:
                return False

            raise

    async def _list(
        self,
        url: str,
        filter_model: TypeAdapter[Any],
        filters: Any,
        paginate: bool = True,
        page_size: Optional[int] = None,
        postprocessor: Optional[PostProcessor] = None,
    ) -> AsyncIterator[Any]:
        filter_data = self._transform_data(filter_model, filters)
        query = self._encode_query_values(filter_data)

        if paginate:
            results = self._paginate(url, query=query, page_size=page_size)
            async for result in results:
                yield result
        else:
            response = await self._session.get(url, query=query)
            response_data = postprocessor(response) if postprocessor else response.json()
            if not isinstance(response_data, list):
                raise ValueError(f"Unexpected response format: {response_data}")

            for result in response_data:
                yield result

    async def _get_by_id(self, url: str, rid: ResourceId) -> Optional[Any]:
        try:
            response = await self._session.get(f"{url}/{rid}")
            return response.json()
        except NcaeHttpError as exc:
            if exc.status_code == 404:
                return None

            raise

    async def _get_by_filters(self, url: str, filter_model: TypeAdapter[Any], filters: Any) -> Optional[Any]:
        results = self._list(url, filter_model, filters, page_size=2)

        # Return early if there are no results, indicating a lack of existence
        first = await next_async(results, None)
        if first is None:
            return None

        # Ensure uniqueness by raising if there is a second result
        second = await next_async(results, None)
        if second is not None:
            raise ValueError(f"Received more than one result for [{url}] with filters: {filters}")

        return first

    async def _paginate(
        self,
        url: str,
        *,
        page_size: Optional[int] = None,
        postprocessor: Optional[PostProcessor] = None,
        **kwargs: Unpack[RequestArgs],
    ) -> AsyncIterator[Any]:
        offset: int = 0
        limit: int = page_size or self.DEFAULT_PAGE_SIZE
        kwargs.setdefault("query", {})

        # Iterate over all available pages
        while True:
            # Update query parameters with pagination details
            kwargs["query"].update({"offset": str(offset), "limit": str(limit)})

            # Fetch current page and extract content as JSON
            response = await self._session.get(url, **kwargs)
            response_data = postprocessor(response) if postprocessor else response.json()

            # Ensure page matches expected structure
            if not isinstance(response_data, dict):
                raise ValueError(f"Unexpected response format: {response_data}")
            if not all(key in response_data for key in {"count", "next", "results"}):
                raise ValueError(f"Missing expected keys in response: {response_data}")

            # Yield current results and break if no further pages are available
            for result in response_data["results"]:
                yield result
            if not response_data["count"] or not response_data["results"] or not response_data["next"]:
                break

            # Update query parameters for next page
            offset += limit

    @classmethod
    def _transform_data(cls, filter_model: TypeAdapter[Any], data: Any) -> Any:
        instance = filter_model.validate_python(data)
        return filter_model.dump_python(instance, by_alias=True, mode="json", context={"external": True})

    @classmethod
    def _encode_query_values(cls, values: dict[str, Any]) -> dict[str, Union[str, list[str]]]:
        query: dict[str, Union[str, list[str]]] = {}
        for key, value in values.items():
            encoded_key = str(key)
            encoded_value = cls._encode_query_value(value)
            if isinstance(encoded_value, list):
                query[f"{encoded_key}[]"] = encoded_value
            else:
                query[encoded_key] = encoded_value

        return query

    @classmethod
    def _encode_query_value(cls, value: Any) -> Union[str, list[str]]:
        if isinstance(value, str):
            return value
        elif isinstance(value, (int, float)):
            return str(value)
        elif isinstance(value, bool):
            return "true" if value else "false"
        elif isinstance(value, list):
            return [str(v) for v in value]
        elif value is None:
            return "null"
        else:
            raise TypeError(f"Unsupported query value type: {type(value)}")



# @generated
# This file was automatically generated by the ncae-sdk code generator.
# DO NOT EDIT MANUALLY, ANY CHANGES WILL BE OVERWRITTEN.

from collections.abc import Iterator
from datetime import datetime
from typing import Any, Final, List, Optional

from typing_extensions import Unpack

from ncae_sdk._resource import ResourceId
from ncae_sdk._sync.endpoint import Endpoint
from ncae_sdk._util import map_sync
from ncae_sdk.resources._schema import (
    InfluxRepository,
    InfluxRepositoryCreate,
    InfluxRepositoryCreateModel,
    InfluxRepositoryFilter,
    InfluxRepositoryFilterModel,
    InfluxRepositoryUpdate,
    InfluxRepositoryUpdateModel,
)
from ncae_sdk.types._models import InfluxDataPoint, InfluxTagFilter


class InfluxRepositoryEndpoint(Endpoint):
    BASE_PATH: Final[str] = "observability/v1/influx-repository"

    def list(self, **filters: Unpack[InfluxRepositoryFilter]) -> Iterator[InfluxRepository]:
        results = self._list(self.BASE_PATH, InfluxRepositoryFilterModel, filters)
        return map_sync(InfluxRepository.parse_api, results)

    def find(self, **filters: Unpack[InfluxRepositoryFilter]) -> Optional[InfluxRepository]:
        result = self._get_by_filters(self.BASE_PATH, InfluxRepositoryFilterModel, filters)
        return InfluxRepository.parse_api(result) if result else None

    def get(self, rid: ResourceId) -> Optional[InfluxRepository]:
        result = self._get_by_id(self.BASE_PATH, rid)
        return InfluxRepository.parse_api(result) if result else None

    def create(self, **payload: Unpack[InfluxRepositoryCreate]) -> InfluxRepository:
        result = self._create(self.BASE_PATH, InfluxRepositoryCreateModel, payload)
        return InfluxRepository.parse_api(result)

    def update(self, rid: ResourceId, /, **payload: Unpack[InfluxRepositoryUpdate]) -> InfluxRepository:
        result = self._update(self.BASE_PATH, rid, InfluxRepositoryUpdateModel, payload)
        return InfluxRepository.parse_api(result)

    def delete(self, rid: ResourceId) -> bool:
        return self._delete(self.BASE_PATH, rid)

    def ingest_metrics(self, rid: ResourceId, points: List[InfluxDataPoint]) -> None:
        points_data = [point.model_dump(exclude_defaults=True, mode="json") for point in points]
        self._session.post(
            f"{self.BASE_PATH}/{rid}/metrics/ingest",
            json={"points": points_data},
        )

    def query_metrics(
        self,
        rid: ResourceId,
        date_from: Optional[datetime] = None,
        date_to: Optional[datetime] = None,
        tag_filters: Optional[List[InfluxTagFilter]] = None,
    ) -> Any:
        date_from_data = date_from.isoformat() if date_from else None
        date_to_data = date_to.isoformat() if date_to else None
        tag_filter_data = (
            [tag_filter.model_dump(exclude_defaults=True, mode="json") for tag_filter in tag_filters]
            if tag_filters
            else []
        )

        response = self._session.post(
            f"{self.BASE_PATH}/{rid}/metrics/query",
            json={
                "date_from": date_from_data,
                "date_to": date_to_data,
                "tag_filters": tag_filter_data,
            },
        )
        return response.json()



from fastapi import FastAPI
from fastapi.exceptions import RequestValidationError

from ncae_sdk.fastapi import _handlers as handlers
from ncae_sdk.fastapi import _middlewares as middlewares
from ncae_sdk.fastapi._config import get_settings
from ncae_sdk.fastapi._lifespan import lifespan
from ncae_sdk.fastapi._logging import setup_default_logging
from ncae_sdk.fastapi._router import system_router


def create_module_app(
    title: str,
    version: str = "0.1.0",
) -> FastAPI:
    settings = get_settings()
    setup_default_logging(settings.debug)

    app = FastAPI(
        title=title,
        version=version,
        lifespan=lifespan,
        openapi_url="/api/schema.json",
        docs_url="/api/docs",
        redoc_url=None,
    )
    app.state.module_name = title
    app.state.module_version = version

    app.include_router(system_router)
    app.exception_handler(RequestValidationError)(handlers.validation_exception_handler)
    app.middleware("http")(middlewares.add_tracing_headers)
    app.middleware("http")(middlewares.inject_request_logger)
    app.middleware("http")(middlewares.log_http_request)

    return app



from collections.abc import AsyncIterator
from typing import Final, Optional

from typing_extensions import Unpack

from ncae_sdk._async.endpoint import AsyncEndpoint
from ncae_sdk._resource import ResourceId
from ncae_sdk._util import map_async
from ncae_sdk.resources._schema import (
    ExtApiAuth,
    ExtApiAuthCreate,
    ExtApiAuthCreateModel,
    ExtApiAuthFilter,
    ExtApiAuthFilterModel,
    ExtApiAuthUpdate,
    ExtApiAuthUpdateModel,
)


class AsyncExtApiAuthEndpoint(AsyncEndpoint):
    BASE_PATH: Final[str] = "automation/v1/auth"

    async def list(self, **filters: Unpack[ExtApiAuthFilter]) -> AsyncIterator[ExtApiAuth]:
        results = self._list(self.BASE_PATH, ExtApiAuthFilterModel, filters)
        return map_async(ExtApiAuth.parse_api, results)

    async def find(self, **filters: Unpack[ExtApiAuthFilter]) -> Optional[ExtApiAuth]:
        result = await self._get_by_filters(self.BASE_PATH, ExtApiAuthFilterModel, filters)
        return ExtApiAuth.parse_api(result) if result else None

    async def get(self, rid: ResourceId) -> Optional[ExtApiAuth]:
        result = await self._get_by_id(self.BASE_PATH, rid)
        return ExtApiAuth.parse_api(result) if result else None

    async def create(self, **payload: Unpack[ExtApiAuthCreate]) -> ExtApiAuth:
        result = await self._create(self.BASE_PATH, ExtApiAuthCreateModel, payload)
        return ExtApiAuth.parse_api(result)

    async def update(self, rid: ResourceId, /, **payload: Unpack[ExtApiAuthUpdate]) -> ExtApiAuth:
        result = await self._update(self.BASE_PATH, rid, ExtApiAuthUpdateModel, payload)
        return ExtApiAuth.parse_api(result)

    async def delete(self, rid: ResourceId) -> bool:
        return await self._delete(self.BASE_PATH, rid)



from typing import Annotated, Any, List, Optional

from pydantic import BeforeValidator, PlainSerializer, WithJsonSchema, WrapSerializer
from pydantic_core.core_schema import SerializerFunctionWrapHandler
from typing_extensions import TypeAlias


def empty_str_to_none(value: Optional[str]) -> Optional[str]:
    if value is None or value == "":
        return None
    return value


def none_to_empty_str(value: Optional[str], nxt: SerializerFunctionWrapHandler) -> str:
    result = nxt(value)
    return "" if result is None else result


def _parse_cmdb_text_list(value: Any) -> list[str]:
    if value is None:
        return []

    if isinstance(value, list):
        return value

    if isinstance(value, str):
        return value.strip().splitlines()

    raise ValueError(f"Expected a string or list of strings, got {type(value)}")


def _serialize_cmdb_text_list(value: list[str]) -> str:
    return "\n".join(value)


CmdbTextList: TypeAlias = Annotated[
    List[str],
    BeforeValidator(_parse_cmdb_text_list),
    PlainSerializer(_serialize_cmdb_text_list),
    WithJsonSchema({"type": "string"}, mode="validation"),
]
CmdbTextList.__doc__ = """
Custom Pydantic field type for handling CMDB fields of type `TextListInput`.
This will always return a list of strings, one per each line in the input field.
While empty lines are not filtered out, a fully empty field will return an empty list.

:type: list[str]
"""

CmdbOptionalStr: TypeAlias = Annotated[
    Optional[str],
    BeforeValidator(empty_str_to_none),
    WrapSerializer(none_to_empty_str),
]
CmdbOptionalStr.__doc__ = """
Custom Pydantic field type for handling optional strings, e.g. any CMDB field of type `TextInput`.
When receiving a value from NCAE Core, it will convert empty strings to `None`.
When sending a value to NCAE Core, it will convert `None` back to an empty string.

:type: Optional[str]
"""



from collections.abc import AsyncIterator, Iterator, MutableMapping
from functools import cached_property
from typing import IO, Any, Final, Mapping, Optional, Union

from httpx import Request as HttpxRequest
from httpx import Response as HttpxResponse
from typing_extensions import NotRequired, TypedDict


class RequestArgs(TypedDict):
    data: NotRequired[Mapping[str, Any]]
    files: NotRequired[Mapping[str, IO[bytes]]]
    headers: NotRequired[MutableMapping[str, str]]
    query: NotRequired[MutableMapping[str, Union[str, list[str]]]]
    json: NotRequired[Any]


class Request:
    def __init__(self, _native: HttpxRequest) -> None:
        self._native: Final[HttpxRequest] = _native

    @property
    def url(self) -> str:
        return str(self._native.url)

    @property
    def query_params(self) -> dict[str, str]:
        return dict(self._native.url.params.items())


class Response:
    def __init__(self, _native: HttpxResponse) -> None:
        self.native: Final[HttpxResponse] = _native

    @property
    def is_success(self) -> bool:
        return self.native.is_success

    @property
    def status_code(self) -> int:
        return self.native.status_code

    @property
    def content(self) -> bytes:
        return self.native.content

    @property
    def text(self) -> str:
        return self.native.text

    @cached_property
    def request(self) -> Request:
        return Request(self.native.request)

    def json(self, **kwargs: Any) -> Any:
        return self.native.json(**kwargs)

    def iter_bytes(self, chunk_size: Optional[int] = None) -> Iterator[bytes]:
        return self.native.iter_bytes(chunk_size=chunk_size)

    async def aiter_bytes(self, chunk_size: Optional[int] = None) -> AsyncIterator[bytes]:
        return self.native.aiter_bytes(chunk_size=chunk_size)

    def read(self) -> bytes:
        return self.native.read()

    async def aread(self) -> bytes:
        return await self.native.aread()



import functools
from traceback import TracebackException
from typing import Any, Awaitable, Callable

from fastapi import BackgroundTasks
from typing_extensions import Concatenate

from ncae_sdk._types import (
    FuncP,
    FuncR,
    inject_parameter,
    process_function_parameters,
    resolve_generic_signatures,
)
from ncae_sdk.fastapi._context import _run_async_or_sync_function
from ncae_sdk.fastapi._dependencies import Router, annotate_fastapi_dependencies
from ncae_sdk.fastapi.extapi._context import ExtApiPhaseContext
from ncae_sdk.types import PhaseStatus


def extapi_phase_route(
    router: Router, path: str
) -> Callable[
    [Callable[FuncP, Any]],
    Callable[Concatenate[ExtApiPhaseContext, BackgroundTasks, FuncP], Awaitable[None]],
]:
    def decorator(
        func: Callable[FuncP, Any],
    ) -> Callable[
        Concatenate[ExtApiPhaseContext, BackgroundTasks, FuncP],
        Awaitable[None],
    ]:
        @router.post(path, status_code=204)
        @process_function_parameters(
            resolve_generic_signatures,
            inject_parameter("__sdk_tasks", BackgroundTasks),
            inject_parameter("__sdk_context", ExtApiPhaseContext),
            annotate_fastapi_dependencies,
        )
        @_wrap_extapi_phase_handler
        @functools.wraps(func)
        async def wrapper(*args: FuncP.args, **kwargs: FuncP.kwargs) -> None:
            await _run_async_or_sync_function(func, *args, **kwargs)

        return wrapper

    return decorator


def extapi_report_route(
    router: Router, path: str
) -> Callable[
    [Callable[FuncP, FuncR]],
    Callable[FuncP, Awaitable[FuncR]],
]:
    def decorator(func: Callable[FuncP, FuncR]) -> Callable[FuncP, Awaitable[FuncR]]:
        @router.post(path)
        @process_function_parameters(
            resolve_generic_signatures,
            annotate_fastapi_dependencies,
        )
        @functools.wraps(func)
        async def wrapper(*args: FuncP.args, **kwargs: FuncP.kwargs) -> FuncR:
            return await _run_async_or_sync_function(func, *args, **kwargs)

        return wrapper

    return decorator


def _wrap_extapi_phase_handler(
    func: Callable[FuncP, Awaitable[None]],
) -> Callable[Concatenate[ExtApiPhaseContext, BackgroundTasks, FuncP], Awaitable[None]]:
    @functools.wraps(func)
    async def wrapper(
        __sdk_context: ExtApiPhaseContext,
        __sdk_tasks: BackgroundTasks,
        *args: FuncP.args,
        **kwargs: FuncP.kwargs,
    ) -> None:
        # Wrapper function to run the actual phase task in the background with basic error handling
        async def phase_task(*task_args: FuncP.args, **task_kwargs: FuncP.kwargs) -> None:
            # Catch and report exceptions during phase execution
            try:
                await func(*task_args, **task_kwargs)
            except Exception as exc:
                tb = TracebackException.from_exception(exc).format()
                await __sdk_context.async_log_error(f"Execution of {func.__name__} failed:\n{''.join(tb)}")
                await __sdk_context.async_update_phase_status(PhaseStatus.ERRORED)
                return

            # Set phase status by default if still stuck on ORDERED
            phase_status = await __sdk_context.async_get_phase_status()
            if phase_status in [PhaseStatus.ORDERED, PhaseStatus.UPDATING, PhaseStatus.RETIRING]:
                if __sdk_context.is_decommission:
                    await __sdk_context.async_update_phase_status(PhaseStatus.RETIRED)
                else:
                    await __sdk_context.async_update_phase_status(PhaseStatus.DEPLOYED)

        __sdk_tasks.add_task(phase_task, *args, **kwargs)

    return wrapper



from collections.abc import AsyncIterator
from concurrent.futures import ThreadPoolExecutor
from contextlib import asynccontextmanager
from typing import TypedDict

from fastapi import FastAPI

from ncae_sdk._async.client import AsyncClient
from ncae_sdk._auth import SessionAuth
from ncae_sdk._sync.client import Client
from ncae_sdk.fastapi._config import get_settings


class LifespanState(TypedDict):
    ncae_client_sync: Client
    ncae_client_async: AsyncClient
    thread_pool: ThreadPoolExecutor


@asynccontextmanager
async def lifespan(_app: FastAPI) -> AsyncIterator[LifespanState]:
    # Startup
    settings = get_settings()
    ncae_auth = SessionAuth(username=settings.ncae_username, password=settings.ncae_password)
    ncae_client_sync = Client(
        base_url=settings.ncae_base_url,
        auth=ncae_auth,
        verify=settings.ncae_verify_tls,
    )
    ncae_client_async = AsyncClient(
        base_url=settings.ncae_base_url,
        auth=ncae_auth,
        verify=settings.ncae_verify_tls,
    )
    thread_pool = ThreadPoolExecutor()

    # Runtime
    state: LifespanState = {
        "ncae_client_sync": ncae_client_sync,
        "ncae_client_async": ncae_client_async,
        "thread_pool": thread_pool,
    }
    yield state

    # Shutdown
    thread_pool.shutdown()
    ncae_client_sync.close()
    await ncae_client_async.close()



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Final

if TYPE_CHECKING:
    from ncae_sdk._async.session import AsyncSession
    from ncae_sdk._sync.session import Session


class Auth(ABC):
    @abstractmethod
    def authenticate(self, session: "Session") -> None:
        """Synchronously authenticate the session using the provided credentials."""
        ...

    @abstractmethod
    async def async_authenticate(self, session: "AsyncSession") -> None:
        """Asynchronously authenticate the session using the provided credentials."""
        ...


class SessionAuth(Auth):
    def __init__(self, username: str, password: str) -> None:
        if not username or not password:
            raise ValueError("Username and password must be provided for authentication.")

        self.username: Final[str] = username
        self.password: Final[str] = password

    def authenticate(self, session: "Session") -> None:
        session.post(
            "auth/v1/login/",
            json={
                "username": self.username,
                "password": self.password,
            },
        )

    async def async_authenticate(self, session: "AsyncSession") -> None:
        await session.post(
            "auth/v1/login/",
            json={
                "username": self.username,
                "password": self.password,
            },
        )



from collections.abc import AsyncIterator
from typing import Final, Optional

from typing_extensions import Unpack

from ncae_sdk._async.endpoint import AsyncEndpoint
from ncae_sdk._resource import ResourceId
from ncae_sdk._util import map_async
from ncae_sdk.resources._schema import (
    Favourite,
    FavouriteCreate,
    FavouriteCreateModel,
    FavouriteFilter,
    FavouriteFilterModel,
    FavouriteUpdate,
    FavouriteUpdateModel,
)


class AsyncFavouriteEndpoint(AsyncEndpoint):
    BASE_PATH: Final[str] = "dashboard/v1/favourite"

    async def list(self, **filters: Unpack[FavouriteFilter]) -> AsyncIterator[Favourite]:
        results = self._list(self.BASE_PATH, FavouriteFilterModel, filters, paginate=False)
        return map_async(Favourite.parse_api, results)

    async def find(self, **filters: Unpack[FavouriteFilter]) -> Optional[Favourite]:
        result = await self._get_by_filters(self.BASE_PATH, FavouriteFilterModel, filters)
        return Favourite.parse_api(result) if result else None

    async def get(self, rid: ResourceId) -> Optional[Favourite]:
        result = await self._get_by_id(self.BASE_PATH, rid)
        return Favourite.parse_api(result) if result else None

    async def create(self, **payload: Unpack[FavouriteCreate]) -> Favourite:
        result = await self._create(self.BASE_PATH, FavouriteCreateModel, payload)
        return Favourite.parse_api(result)

    async def update(self, rid: ResourceId, /, **payload: Unpack[FavouriteUpdate]) -> Favourite:
        result = await self._update(self.BASE_PATH, rid, FavouriteUpdateModel, payload)
        return Favourite.parse_api(result)

    async def delete(self, rid: ResourceId) -> bool:
        return await self._delete(self.BASE_PATH, rid)



from datetime import datetime
from typing import Any, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, computed_field


class Model(BaseModel):
    model_config = ConfigDict(
        extra="ignore",
        serialize_by_alias=True,
        validate_by_alias=True,
        validate_by_name=True,
    )


class AuditLogChange(Model):
    field: str = Field(alias="field")
    old: str = Field(alias="old")
    new: str = Field(alias="new")


class AwxInventoryItem(Model):
    id: int
    host: str = Field(alias="ip")
    name: str
    group_ids: list[int] = Field(alias="groups")
    variables: dict[str, Any]


class InfluxDataPoint(Model):
    measurement: str = Field(alias="measurement")
    timestamp: Optional[datetime] = Field(alias="timestamp", default=None)
    field_data: dict[str, Union[str, int, float, bool]] = Field(alias="field_data")
    tag_data: dict[str, str] = Field(alias="tag_data", default_factory=dict)


class InfluxTagFilter(Model):
    key: str = Field(alias="key")
    value: str = Field(alias="value")

    @computed_field  # type: ignore[prop-decorator]
    @property
    def mode(self) -> str:
        return "exact"



# @generated
# This file was automatically generated by the ncae-sdk code generator.
# DO NOT EDIT MANUALLY, ANY CHANGES WILL BE OVERWRITTEN.

from collections.abc import Iterator
from typing import Final

from typing_extensions import Unpack

from ncae_sdk._sync.endpoint import Endpoint
from ncae_sdk._util import map_sync
from ncae_sdk.resources._schema import (
    UsageLog,
    UsageLogFilter,
    UsageLogFilterModel,
)


class UsageLogEndpoint(Endpoint):
    BASE_PATH: Final[str] = "logger/v1/usage-log"

    def list(self, **filters: Unpack[UsageLogFilter]) -> Iterator[UsageLog]:
        results = self._list(self.BASE_PATH, UsageLogFilterModel, filters)
        return map_sync(UsageLog.parse_api, results)



from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime, timedelta
from threading import Lock
from typing import Awaitable, Callable, DefaultDict, Final, Optional

from typing_extensions import TypeAlias


@dataclass(frozen=True)
class ProgressUpdate:
    elapsed_time: timedelta
    percentage: float
    current: int
    total: int


class ProgressTracker:
    STEP_MIN_PERCENTAGE_CHANGE: Final[int] = 1
    STEP_MIN_TIME_SINCE_MSECS: Final[int] = 1000

    def __init__(self, expected_total: int):
        self._lock: Final[Lock] = Lock()
        self._store: Final[DefaultDict[str, int]] = defaultdict(int)
        self._expected_total: Final[int] = expected_total
        self._last_update: datetime = datetime.now()
        self._last_percentage: float = 0
        self._start_time: Optional[datetime] = None

    def increment(self, key: str) -> Optional[ProgressUpdate]:
        with self._lock:
            self._store[key] += 1

        return self._check_for_update()

    def reset(self) -> None:
        with self._lock:
            self._store.clear()

    def _check_for_update(self) -> Optional[ProgressUpdate]:
        # Set start time on first increment
        if self._start_time is None:
            self._start_time = datetime.now()

        # Calculate current percentage and time since last update
        percentage = (self.current_total / self.expected_total) * 100
        time_delta = datetime.now() - self._last_update

        # Determine if eligible for progress update
        has_completed = self.current_total == self.expected_total
        has_enough_time = time_delta.total_seconds() * 1000 >= self.STEP_MIN_TIME_SINCE_MSECS
        has_enough_progress = percentage - self._last_percentage >= self.STEP_MIN_PERCENTAGE_CHANGE

        # Return progress update if either completed or enough progress/time has passed
        if not (has_completed or (has_enough_time and has_enough_progress)):
            return None

        return ProgressUpdate(
            elapsed_time=datetime.now() - self._start_time if self._start_time else timedelta(0),
            percentage=percentage,
            current=self.current_total,
            total=self.expected_total,
        )

    def __str__(self) -> str:
        return ", ".join(f"{key}={count}" for key, count in self._store.items())

    @property
    def current_total(self) -> int:
        return sum(self._store.values())

    @property
    def expected_total(self) -> int:
        return self._expected_total


ProgressHookSync: TypeAlias = Callable[[ProgressUpdate], None]
ProgressHookAsync: TypeAlias = Callable[[ProgressUpdate], Awaitable[None]]



# @generated
# This file was automatically generated by the ncae-sdk code generator.
# DO NOT EDIT MANUALLY, ANY CHANGES WILL BE OVERWRITTEN.
from ncae_sdk.resources._schema import (
    AuditLog,
    AuditLogFilter,
    AwxModule,
    AwxModuleCreate,
    AwxModuleFilter,
    AwxModuleUpdate,
    BasePhase,
    BasePhaseFilter,
    ChangeSchedule,
    ChangeScheduleCreate,
    ChangeScheduleFilter,
    ChangeScheduleRun,
    ChangeScheduleRunCreate,
    ChangeScheduleRunFilter,
    ChangeScheduleRunUpdate,
    ChangeScheduleUpdate,
    CmdbEntry,
    CmdbEntryCreate,
    CmdbEntryFilter,
    CmdbEntryUpdate,
    Credential,
    CredentialCreate,
    CredentialFilter,
    CredentialUpdate,
    Device,
    DeviceCreate,
    DeviceFilter,
    DeviceGroup,
    DeviceGroupCreate,
    DeviceGroupFilter,
    DeviceGroupUpdate,
    DeviceModel,
    DeviceModelFilter,
    DeviceUpdate,
    ExtApiAuth,
    ExtApiAuthCreate,
    ExtApiAuthFilter,
    ExtApiAuthUpdate,
    ExtApiPhase,
    ExtApiPhaseCreate,
    ExtApiPhaseFilter,
    ExtApiPhaseUpdate,
    ExtApiService,
    ExtApiServiceCreate,
    ExtApiServiceFilter,
    ExtApiServiceUpdate,
    Favourite,
    FavouriteCreate,
    FavouriteFilter,
    FavouriteUpdate,
    GrafanaDashboard,
    GrafanaDashboardCreate,
    GrafanaDashboardFilter,
    GrafanaDashboardUpdate,
    GrafanaSystem,
    GrafanaSystemCreate,
    GrafanaSystemFilter,
    GrafanaSystemUpdate,
    InfluxRepository,
    InfluxRepositoryCreate,
    InfluxRepositoryFilter,
    InfluxRepositoryUpdate,
    InfluxSystem,
    InfluxSystemCreate,
    InfluxSystemFilter,
    InfluxSystemUpdate,
    Module,
    ModuleFilter,
    Phase,
    PhaseCreate,
    PhaseFilter,
    PhaseInstance,
    PhaseInstanceCreate,
    PhaseInstanceFilter,
    PhaseInstanceUpdate,
    PhaseUpdate,
    PollingPhase,
    PollingPhaseCreate,
    PollingPhaseFilter,
    PollingPhaseUpdate,
    PythonScriptPhase,
    PythonScriptPhaseCreate,
    PythonScriptPhaseFilter,
    PythonScriptPhaseUpdate,
    Report,
    ReportCreate,
    ReportFilter,
    ReportUpdate,
    Service,
    ServiceCreate,
    ServiceFilter,
    ServiceInstance,
    ServiceInstanceCreate,
    ServiceInstanceFilter,
    ServiceInstanceLog,
    ServiceInstanceLogCreate,
    ServiceInstanceLogFilter,
    ServiceInstanceLogUpdate,
    ServiceInstanceUpdate,
    ServiceManagementPhase,
    ServiceManagementPhaseCreate,
    ServiceManagementPhaseFilter,
    ServiceManagementPhaseUpdate,
    ServiceUpdate,
    SharedDataBucket,
    SharedDataBucketCreate,
    SharedDataBucketFilter,
    SharedDataBucketUpdate,
    SharedDataSubscriber,
    SharedDataSubscriberCreate,
    SharedDataSubscriberFilter,
    SharedDataSubscriberUpdate,
    Tag,
    TagCreate,
    TagFilter,
    TagRelation,
    TagRelationCreate,
    TagRelationFilter,
    TagRelationUpdate,
    TagUpdate,
    Tenant,
    TenantFilter,
    Transaction,
    TransactionCreate,
    TransactionFilter,
    TransactionUpdate,
    UploadFile,
    UploadFileCreate,
    UploadFileFilter,
    UploadFileUpdate,
    UsageLog,
    UsageLogFilter,
    User,
    UserCreate,
    UserFilter,
    UserUpdate,
)

__all__ = [
    "AuditLog",
    "AuditLogFilter",
    "AwxModule",
    "AwxModuleCreate",
    "AwxModuleUpdate",
    "AwxModuleFilter",
    "BasePhase",
    "BasePhaseFilter",
    "ChangeSchedule",
    "ChangeScheduleCreate",
    "ChangeScheduleUpdate",
    "ChangeScheduleFilter",
    "ChangeScheduleRun",
    "ChangeScheduleRunCreate",
    "ChangeScheduleRunUpdate",
    "ChangeScheduleRunFilter",
    "CmdbEntry",
    "CmdbEntryCreate",
    "CmdbEntryUpdate",
    "CmdbEntryFilter",
    "Credential",
    "CredentialCreate",
    "CredentialUpdate",
    "CredentialFilter",
    "Device",
    "DeviceCreate",
    "DeviceUpdate",
    "DeviceFilter",
    "DeviceGroup",
    "DeviceGroupCreate",
    "DeviceGroupUpdate",
    "DeviceGroupFilter",
    "DeviceModel",
    "DeviceModelFilter",
    "ExtApiAuth",
    "ExtApiAuthCreate",
    "ExtApiAuthUpdate",
    "ExtApiAuthFilter",
    "ExtApiPhase",
    "ExtApiPhaseCreate",
    "ExtApiPhaseUpdate",
    "ExtApiPhaseFilter",
    "ExtApiService",
    "ExtApiServiceCreate",
    "ExtApiServiceUpdate",
    "ExtApiServiceFilter",
    "Favourite",
    "FavouriteCreate",
    "FavouriteUpdate",
    "FavouriteFilter",
    "GrafanaDashboard",
    "GrafanaDashboardCreate",
    "GrafanaDashboardUpdate",
    "GrafanaDashboardFilter",
    "GrafanaSystem",
    "GrafanaSystemCreate",
    "GrafanaSystemUpdate",
    "GrafanaSystemFilter",
    "InfluxRepository",
    "InfluxRepositoryCreate",
    "InfluxRepositoryUpdate",
    "InfluxRepositoryFilter",
    "InfluxSystem",
    "InfluxSystemCreate",
    "InfluxSystemUpdate",
    "InfluxSystemFilter",
    "Module",
    "ModuleFilter",
    "Phase",
    "PhaseCreate",
    "PhaseUpdate",
    "PhaseFilter",
    "PhaseInstance",
    "PhaseInstanceCreate",
    "PhaseInstanceUpdate",
    "PhaseInstanceFilter",
    "PollingPhase",
    "PollingPhaseCreate",
    "PollingPhaseUpdate",
    "PollingPhaseFilter",
    "PythonScriptPhase",
    "PythonScriptPhaseCreate",
    "PythonScriptPhaseUpdate",
    "PythonScriptPhaseFilter",
    "Report",
    "ReportCreate",
    "ReportUpdate",
    "ReportFilter",
    "Service",
    "ServiceCreate",
    "ServiceUpdate",
    "ServiceFilter",
    "ServiceInstance",
    "ServiceInstanceCreate",
    "ServiceInstanceUpdate",
    "ServiceInstanceFilter",
    "ServiceInstanceLog",
    "ServiceInstanceLogCreate",
    "ServiceInstanceLogUpdate",
    "ServiceInstanceLogFilter",
    "ServiceManagementPhase",
    "ServiceManagementPhaseCreate",
    "ServiceManagementPhaseUpdate",
    "ServiceManagementPhaseFilter",
    "SharedDataBucket",
    "SharedDataBucketCreate",
    "SharedDataBucketUpdate",
    "SharedDataBucketFilter",
    "SharedDataSubscriber",
    "SharedDataSubscriberCreate",
    "SharedDataSubscriberUpdate",
    "SharedDataSubscriberFilter",
    "Tag",
    "TagCreate",
    "TagUpdate",
    "TagFilter",
    "TagRelation",
    "TagRelationCreate",
    "TagRelationUpdate",
    "TagRelationFilter",
    "Tenant",
    "TenantFilter",
    "Transaction",
    "TransactionCreate",
    "TransactionUpdate",
    "TransactionFilter",
    "UploadFile",
    "UploadFileCreate",
    "UploadFileUpdate",
    "UploadFileFilter",
    "UsageLog",
    "UsageLogFilter",
    "User",
    "UserCreate",
    "UserUpdate",
    "UserFilter",
]



from collections.abc import AsyncIterator
from typing import Final, Optional

from typing_extensions import Unpack

from ncae_sdk._async.endpoint import AsyncEndpoint
from ncae_sdk._resource import ResourceId
from ncae_sdk._util import map_async
from ncae_sdk.resources._schema import (
    ServiceManagementPhase,
    ServiceManagementPhaseCreate,
    ServiceManagementPhaseCreateModel,
    ServiceManagementPhaseFilter,
    ServiceManagementPhaseFilterModel,
    ServiceManagementPhaseUpdate,
    ServiceManagementPhaseUpdateModel,
)


class AsyncServiceManagementPhaseEndpoint(AsyncEndpoint):
    BASE_PATH: Final[str] = "automation/v1/phase/service-management-phase"

    async def list(self, **filters: Unpack[ServiceManagementPhaseFilter]) -> AsyncIterator[ServiceManagementPhase]:
        results = self._list(self.BASE_PATH, ServiceManagementPhaseFilterModel, filters)
        return map_async(ServiceManagementPhase.parse_api, results)

    async def find(self, **filters: Unpack[ServiceManagementPhaseFilter]) -> Optional[ServiceManagementPhase]:
        result = await self._get_by_filters(self.BASE_PATH, ServiceManagementPhaseFilterModel, filters)
        return ServiceManagementPhase.parse_api(result) if result else None

    async def get(self, rid: ResourceId) -> Optional[ServiceManagementPhase]:
        result = await self._get_by_id(self.BASE_PATH, rid)
        return ServiceManagementPhase.parse_api(result) if result else None

    async def create(self, **payload: Unpack[ServiceManagementPhaseCreate]) -> ServiceManagementPhase:
        result = await self._create(self.BASE_PATH, ServiceManagementPhaseCreateModel, payload)
        return ServiceManagementPhase.parse_api(result)

    async def update(
        self, rid: ResourceId, /, **payload: Unpack[ServiceManagementPhaseUpdate]
    ) -> ServiceManagementPhase:
        result = await self._update(self.BASE_PATH, rid, ServiceManagementPhaseUpdateModel, payload)
        return ServiceManagementPhase.parse_api(result)

    async def delete(self, rid: ResourceId) -> bool:
        return await self._delete(self.BASE_PATH, rid)



from __future__ import annotations

from typing import Annotated, Any, Literal, Union

from pydantic import BaseModel, ConfigDict, Field


class BaseTemplate(BaseModel):
    model_config = ConfigDict(
        serialize_by_alias=True,
        validate_by_name=True,
        validate_by_alias=False,
    )
    name: str
    label: str


class HiddenTemplate(BaseTemplate):
    type: Literal["hidden"]
    initial_value: str = Field(alias="initialValue")


class TextTemplate(BaseTemplate):
    type: Literal["text"]
    editable: bool = True
    required: bool = True


class TextListTemplate(BaseTemplate):
    type: Literal["text-list"]
    editable: bool = True
    required: bool = True


class GroupTemplate(BaseTemplate):
    type: Literal["group"]
    children: list["Template"] = Field(default_factory=list)
    editable: bool = True
    required: bool = True

    def model_post_init(self, context: Any, /) -> None:
        identifier = HiddenTemplate(
            type="hidden",
            name="identifier",
            label="identifier",
            initial_value="uuid4",
        )
        self.children.insert(0, identifier)


Template = Annotated[
    Union[HiddenTemplate, TextTemplate, TextListTemplate, GroupTemplate],
    Field(discriminator="type"),
]



# @generated
# This file was automatically generated by the ncae-sdk code generator.
# DO NOT EDIT MANUALLY, ANY CHANGES WILL BE OVERWRITTEN.
from ncae_sdk.endpoints._async.audit_log import AsyncAuditLogEndpoint
from ncae_sdk.endpoints._async.awx_module import AsyncAwxModuleEndpoint
from ncae_sdk.endpoints._async.base_phase import AsyncBasePhaseEndpoint
from ncae_sdk.endpoints._async.change_schedule import AsyncChangeScheduleEndpoint
from ncae_sdk.endpoints._async.change_schedule_run import AsyncChangeScheduleRunEndpoint
from ncae_sdk.endpoints._async.cmdb_entry import AsyncCmdbEntryEndpoint
from ncae_sdk.endpoints._async.credential import AsyncCredentialEndpoint
from ncae_sdk.endpoints._async.device import AsyncDeviceEndpoint
from ncae_sdk.endpoints._async.device_group import AsyncDeviceGroupEndpoint
from ncae_sdk.endpoints._async.device_model import AsyncDeviceModelEndpoint
from ncae_sdk.endpoints._async.ext_api_auth import AsyncExtApiAuthEndpoint
from ncae_sdk.endpoints._async.ext_api_phase import AsyncExtApiPhaseEndpoint
from ncae_sdk.endpoints._async.ext_api_service import AsyncExtApiServiceEndpoint
from ncae_sdk.endpoints._async.favourite import AsyncFavouriteEndpoint
from ncae_sdk.endpoints._async.grafana_dashboard import AsyncGrafanaDashboardEndpoint
from ncae_sdk.endpoints._async.grafana_system import AsyncGrafanaSystemEndpoint
from ncae_sdk.endpoints._async.influx_repository import AsyncInfluxRepositoryEndpoint
from ncae_sdk.endpoints._async.influx_system import AsyncInfluxSystemEndpoint
from ncae_sdk.endpoints._async.module import AsyncModuleEndpoint
from ncae_sdk.endpoints._async.phase import AsyncPhaseEndpoint
from ncae_sdk.endpoints._async.phase_instance import AsyncPhaseInstanceEndpoint
from ncae_sdk.endpoints._async.polling_phase import AsyncPollingPhaseEndpoint
from ncae_sdk.endpoints._async.python_script_phase import AsyncPythonScriptPhaseEndpoint
from ncae_sdk.endpoints._async.report import AsyncReportEndpoint
from ncae_sdk.endpoints._async.service import AsyncServiceEndpoint
from ncae_sdk.endpoints._async.service_instance import AsyncServiceInstanceEndpoint
from ncae_sdk.endpoints._async.service_instance_log import AsyncServiceInstanceLogEndpoint
from ncae_sdk.endpoints._async.service_management_phase import AsyncServiceManagementPhaseEndpoint
from ncae_sdk.endpoints._async.shared_data_bucket import AsyncSharedDataBucketEndpoint
from ncae_sdk.endpoints._async.shared_data_subscriber import AsyncSharedDataSubscriberEndpoint
from ncae_sdk.endpoints._async.tag import AsyncTagEndpoint
from ncae_sdk.endpoints._async.tag_relation import AsyncTagRelationEndpoint
from ncae_sdk.endpoints._async.tenant import AsyncTenantEndpoint
from ncae_sdk.endpoints._async.transaction import AsyncTransactionEndpoint
from ncae_sdk.endpoints._async.upload_file import AsyncUploadFileEndpoint
from ncae_sdk.endpoints._async.usage_log import AsyncUsageLogEndpoint
from ncae_sdk.endpoints._async.user import AsyncUserEndpoint
from ncae_sdk.endpoints._sync.audit_log import AuditLogEndpoint
from ncae_sdk.endpoints._sync.awx_module import AwxModuleEndpoint
from ncae_sdk.endpoints._sync.base_phase import BasePhaseEndpoint
from ncae_sdk.endpoints._sync.change_schedule import ChangeScheduleEndpoint
from ncae_sdk.endpoints._sync.change_schedule_run import ChangeScheduleRunEndpoint
from ncae_sdk.endpoints._sync.cmdb_entry import CmdbEntryEndpoint
from ncae_sdk.endpoints._sync.credential import CredentialEndpoint
from ncae_sdk.endpoints._sync.device import DeviceEndpoint
from ncae_sdk.endpoints._sync.device_group import DeviceGroupEndpoint
from ncae_sdk.endpoints._sync.device_model import DeviceModelEndpoint
from ncae_sdk.endpoints._sync.ext_api_auth import ExtApiAuthEndpoint
from ncae_sdk.endpoints._sync.ext_api_phase import ExtApiPhaseEndpoint
from ncae_sdk.endpoints._sync.ext_api_service import ExtApiServiceEndpoint
from ncae_sdk.endpoints._sync.favourite import FavouriteEndpoint
from ncae_sdk.endpoints._sync.grafana_dashboard import GrafanaDashboardEndpoint
from ncae_sdk.endpoints._sync.grafana_system import GrafanaSystemEndpoint
from ncae_sdk.endpoints._sync.influx_repository import InfluxRepositoryEndpoint
from ncae_sdk.endpoints._sync.influx_system import InfluxSystemEndpoint
from ncae_sdk.endpoints._sync.module import ModuleEndpoint
from ncae_sdk.endpoints._sync.phase import PhaseEndpoint
from ncae_sdk.endpoints._sync.phase_instance import PhaseInstanceEndpoint
from ncae_sdk.endpoints._sync.polling_phase import PollingPhaseEndpoint
from ncae_sdk.endpoints._sync.python_script_phase import PythonScriptPhaseEndpoint
from ncae_sdk.endpoints._sync.report import ReportEndpoint
from ncae_sdk.endpoints._sync.service import ServiceEndpoint
from ncae_sdk.endpoints._sync.service_instance import ServiceInstanceEndpoint
from ncae_sdk.endpoints._sync.service_instance_log import ServiceInstanceLogEndpoint
from ncae_sdk.endpoints._sync.service_management_phase import ServiceManagementPhaseEndpoint
from ncae_sdk.endpoints._sync.shared_data_bucket import SharedDataBucketEndpoint
from ncae_sdk.endpoints._sync.shared_data_subscriber import SharedDataSubscriberEndpoint
from ncae_sdk.endpoints._sync.tag import TagEndpoint
from ncae_sdk.endpoints._sync.tag_relation import TagRelationEndpoint
from ncae_sdk.endpoints._sync.tenant import TenantEndpoint
from ncae_sdk.endpoints._sync.transaction import TransactionEndpoint
from ncae_sdk.endpoints._sync.upload_file import UploadFileEndpoint
from ncae_sdk.endpoints._sync.usage_log import UsageLogEndpoint
from ncae_sdk.endpoints._sync.user import UserEndpoint

__all__ = [
    "AsyncAuditLogEndpoint",
    "AsyncAwxModuleEndpoint",
    "AsyncBasePhaseEndpoint",
    "AsyncChangeScheduleEndpoint",
    "AsyncChangeScheduleRunEndpoint",
    "AsyncCmdbEntryEndpoint",
    "AsyncCredentialEndpoint",
    "AsyncDeviceEndpoint",
    "AsyncDeviceGroupEndpoint",
    "AsyncDeviceModelEndpoint",
    "AsyncExtApiAuthEndpoint",
    "AsyncExtApiPhaseEndpoint",
    "AsyncExtApiServiceEndpoint",
    "AsyncFavouriteEndpoint",
    "AsyncGrafanaDashboardEndpoint",
    "AsyncGrafanaSystemEndpoint",
    "AsyncInfluxRepositoryEndpoint",
    "AsyncInfluxSystemEndpoint",
    "AsyncModuleEndpoint",
    "AsyncPhaseEndpoint",
    "AsyncPhaseInstanceEndpoint",
    "AsyncPollingPhaseEndpoint",
    "AsyncPythonScriptPhaseEndpoint",
    "AsyncReportEndpoint",
    "AsyncServiceEndpoint",
    "AsyncServiceInstanceEndpoint",
    "AsyncServiceInstanceLogEndpoint",
    "AsyncServiceManagementPhaseEndpoint",
    "AsyncSharedDataBucketEndpoint",
    "AsyncSharedDataSubscriberEndpoint",
    "AsyncTagEndpoint",
    "AsyncTagRelationEndpoint",
    "AsyncTenantEndpoint",
    "AsyncTransactionEndpoint",
    "AsyncUploadFileEndpoint",
    "AsyncUsageLogEndpoint",
    "AsyncUserEndpoint",
    "AuditLogEndpoint",
    "AwxModuleEndpoint",
    "BasePhaseEndpoint",
    "ChangeScheduleEndpoint",
    "ChangeScheduleRunEndpoint",
    "CmdbEntryEndpoint",
    "CredentialEndpoint",
    "DeviceEndpoint",
    "DeviceGroupEndpoint",
    "DeviceModelEndpoint",
    "ExtApiAuthEndpoint",
    "ExtApiPhaseEndpoint",
    "ExtApiServiceEndpoint",
    "FavouriteEndpoint",
    "GrafanaDashboardEndpoint",
    "GrafanaSystemEndpoint",
    "InfluxRepositoryEndpoint",
    "InfluxSystemEndpoint",
    "ModuleEndpoint",
    "PhaseEndpoint",
    "PhaseInstanceEndpoint",
    "PollingPhaseEndpoint",
    "PythonScriptPhaseEndpoint",
    "ReportEndpoint",
    "ServiceEndpoint",
    "ServiceInstanceEndpoint",
    "ServiceInstanceLogEndpoint",
    "ServiceManagementPhaseEndpoint",
    "SharedDataBucketEndpoint",
    "SharedDataSubscriberEndpoint",
    "TagEndpoint",
    "TagRelationEndpoint",
    "TenantEndpoint",
    "TransactionEndpoint",
    "UploadFileEndpoint",
    "UsageLogEndpoint",
    "UserEndpoint",
]



# @generated
# This file was automatically generated by the ncae-sdk code generator.
# DO NOT EDIT MANUALLY, ANY CHANGES WILL BE OVERWRITTEN.

from collections.abc import Iterator
from typing import Final, Optional

from typing_extensions import Unpack

from ncae_sdk._resource import ResourceId
from ncae_sdk._sync.endpoint import Endpoint
from ncae_sdk._util import map_sync
from ncae_sdk.resources._schema import (
    ExtApiPhase,
    ExtApiPhaseCreate,
    ExtApiPhaseCreateModel,
    ExtApiPhaseFilter,
    ExtApiPhaseFilterModel,
    ExtApiPhaseUpdate,
    ExtApiPhaseUpdateModel,
)


class ExtApiPhaseEndpoint(Endpoint):
    BASE_PATH: Final[str] = "automation/v1/phase/ext-api-phase"

    def list(self, **filters: Unpack[ExtApiPhaseFilter]) -> Iterator[ExtApiPhase]:
        results = self._list(self.BASE_PATH, ExtApiPhaseFilterModel, filters)
        return map_sync(ExtApiPhase.parse_api, results)

    def find(self, **filters: Unpack[ExtApiPhaseFilter]) -> Optional[ExtApiPhase]:
        result = self._get_by_filters(self.BASE_PATH, ExtApiPhaseFilterModel, filters)
        return ExtApiPhase.parse_api(result) if result else None

    def get(self, rid: ResourceId) -> Optional[ExtApiPhase]:
        result = self._get_by_id(self.BASE_PATH, rid)
        return ExtApiPhase.parse_api(result) if result else None

    def create(self, **payload: Unpack[ExtApiPhaseCreate]) -> ExtApiPhase:
        result = self._create(self.BASE_PATH, ExtApiPhaseCreateModel, payload)
        return ExtApiPhase.parse_api(result)

    def update(self, rid: ResourceId, /, **payload: Unpack[ExtApiPhaseUpdate]) -> ExtApiPhase:
        result = self._update(self.BASE_PATH, rid, ExtApiPhaseUpdateModel, payload)
        return ExtApiPhase.parse_api(result)

    def delete(self, rid: ResourceId) -> bool:
        return self._delete(self.BASE_PATH, rid)



from abc import ABC, abstractmethod
from typing import Annotated, Any, Optional, TypeVar, Union, cast

from pydantic import BaseModel, BeforeValidator, ConfigDict, WrapSerializer, with_config
from pydantic_core import PydanticUndefined
from pydantic_core.core_schema import SerializationInfo, SerializerFunctionWrapHandler
from typing_extensions import Self, TypeAlias, TypedDict

T = TypeVar("T")

ResourceId = Union[int, str]
ResourceT = TypeVar("ResourceT", bound="Resource")


class Schema(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
        frozen=True,
        serialize_by_alias=True,
        validate_by_alias=False,
        validate_by_name=True,
    )


class Resource(BaseModel, ABC):
    model_config = ConfigDict(
        extra="ignore",
        frozen=True,
        validate_by_alias=True,
        validate_by_name=True,
    )

    @classmethod
    def parse_api(cls, raw: Any) -> Self:
        return cls.model_validate(raw)

    @staticmethod
    @abstractmethod
    def get_schema() -> type[Schema]: ...

    @staticmethod
    @abstractmethod
    def get_filter_schema() -> type[Schema]: ...


@with_config(
    extra="forbid",
    frozen=True,
    validate_by_alias=False,
    validate_by_name=True,
)
class ResourceDict(TypedDict):
    pass


def is_external_serialization(info: SerializationInfo) -> bool:
    return isinstance(info.context, dict) and info.context.get("external", False)


def empty_str_to_none(value: Optional[str]) -> Optional[str]:
    if value is None or value == "":
        return None
    return value


def none_to_empty_str(
    value: Optional[str],
    handler: SerializerFunctionWrapHandler,
    info: SerializationInfo,
) -> Optional[str]:
    result = handler(value)
    if not is_external_serialization(info):
        return cast(Optional[str], result)

    return "" if result is None else result


def extract_nested_id(value: Any) -> Any:
    if isinstance(value, int) or isinstance(value, str):
        return value
    if isinstance(value, dict) and "id" in value:
        return value["id"]
    return PydanticUndefined


def serialize_nested_id(
    value: Any,
    handler: SerializerFunctionWrapHandler,
    info: SerializationInfo,
) -> Union[int, dict[str, Any]]:
    result = handler(value)
    if not is_external_serialization(info):
        return cast(int, result)

    return {"id": result}


OptionalStr: TypeAlias = Annotated[
    Optional[str],
    BeforeValidator(empty_str_to_none),
    WrapSerializer(none_to_empty_str),
]

NestedId: TypeAlias = Annotated[T, BeforeValidator(extract_nested_id), WrapSerializer(serialize_nested_id)]
ReadOnlyNestedId: TypeAlias = Annotated[T, BeforeValidator(extract_nested_id)]



from dataclasses import dataclass
from functools import cached_property
from typing import Optional

from pydantic import computed_field
from typing_extensions import Self


@dataclass(frozen=True)
class SessionContext:
    ncae_tenant_id: Optional[int] = None
    ncae_transaction_id: Optional[str] = None
    extra_headers: Optional[dict[str, str]] = None

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def request_headers(self) -> dict[str, str]:
        headers: dict[str, str] = {}
        if self.ncae_tenant_id is not None:
            headers["X-Tenant-Id"] = str(self.ncae_tenant_id)
        if self.ncae_transaction_id is not None:
            headers["X-Transaction-Id"] = self.ncae_transaction_id
        if self.extra_headers is not None:
            headers.update(self.extra_headers)

        return headers

    @classmethod
    def empty(cls) -> Self:
        return cls()



from typing import Generic, Optional

from typing_extensions import Protocol

from ncae_sdk.fastapi.extapi._contracts import CallbackT, CmdbT, ExtApiPhaseExtraVars


class ExtApiPhaseContextProtocol(Protocol, Generic[CmdbT, CallbackT]):
    @property
    def callback_data(self) -> CallbackT: ...

    @property
    def cmdb_data(self) -> CmdbT: ...

    @property
    def cmdb_data_previous(self) -> Optional[CmdbT]: ...

    @property
    def body(self) -> ExtApiPhaseExtraVars[CmdbT, CallbackT]: ...
